#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math
from typing import List, Tuple, Type

import numpy as np
import pytest
import torch
from flsim.clients.base_client import ClientConfig
from flsim.common.pytest_helper import (
    assertAlmostEqual,
    assertEqual,
    assertTrue,
    assertLessEqual,
    assertGreaterEqual,
)
from flsim.common.training_event_handler import TestAsyncTrainingEventHandler
from flsim.common.training_simulator import AsyncTrainingSimulator
from flsim.data.data_provider import FLDataProviderFromList, IFLDataProvider
from flsim.tests.utils import check_inherit_logging_level
from flsim.utils.async_trainer.async_user_selector import (
    AsyncUserSelector,
    RandomAsyncUserSelector,
    RoundRobinAsyncUserSelector,
)
from flsim.utils.async_trainer.device_state import DeviceState, TrainingState
from flsim.utils.async_trainer.training_event_generator import (
    AsyncTrainingEventGenerator,
    AsyncTrainingEventGeneratorConfig,
    AsyncTrainingEventGeneratorFromList,
    AsyncTrainingEventGeneratorFromListConfig,
    ConstantAsyncTrainingStartTimeDistrConfig,
    EventTimingInfo,
    PoissonAsyncTrainingStartTimeDistrConfig,
)
from flsim.utils.sample_model import MockFLModel
from flsim.utils.tests.helpers.test_training_simulator_utils import (  # noqa: B950 line is too long
    ConstantEventGenTestConfig,
    ConstantEventGenTestConfigPerUserGaussian,
    PoissonEventGenTestConfig,
    PoissonEventGenTestConfigPerUserGaussian,
)
from flsim.utils.timing.training_duration_distribution import (
    DurationDistributionConfig,
    PerExampleGaussianDurationDistributionConfig,
    PerUserGaussianDurationDistributionConfig,
    PerUserHalfNormalDurationDistributionConfig,
    PerUserUniformDurationDistributionConfig,
    PerUserExponentialDurationDistributionConfig,
)
from omegaconf import OmegaConf


class Globals:
    DEVICE_STATE_IDX = 0
    TRAINING_STATE_IDX = 1


@pytest.fixture(scope="class")
def prepare_training_simulator_test_utils(request):
    request.cls.shared_client_config = ClientConfig(
        epochs=1,
        max_clip_norm_normalized=0,
        only_federated_params=True,
        random_seed=1,
        store_models_and_optimizers=False,
    )


@pytest.mark.usefixtures("prepare_training_simulator_test_utils")
class TestTrainingSimulatorUtils:
    def random_user_selector(self, data_provider: IFLDataProvider) -> AsyncUserSelector:
        return RandomAsyncUserSelector(data_provider=data_provider)

    def round_robin_user_selector(
        self, data_provider: IFLDataProvider
    ) -> AsyncUserSelector:
        return RoundRobinAsyncUserSelector(data_provider=data_provider)

    def _create_data_provider(self, num_users: int, examples_per_user: int):
        # one_user_data has 1 batch of len = examples_per_user
        one_user_data = [list(range(examples_per_user))]
        data = [one_user_data] * num_users

        return FLDataProviderFromList(
            train_user_list=data,
            eval_user_list=data,
            test_user_list=data,
            model=MockFLModel(num_examples_per_user=examples_per_user),
        )

    def _verify_event(
        self,
        training_event: Tuple[DeviceState, TrainingState],
        exp_start_time: int,
        exp_end_time: int,
        exp_training_state: TrainingState,
    ):
        assertEqual(
            # pyre-fixme[16]: `TrainingState` has no attribute `training_start_time`.
            training_event[Globals.DEVICE_STATE_IDX].training_schedule.start_time,
            exp_start_time,
        )
        assertEqual(
            training_event[Globals.DEVICE_STATE_IDX].training_schedule.end_time,
            exp_end_time,
        )
        assertEqual(training_event[Globals.TRAINING_STATE_IDX], exp_training_state)

    def test_training_simulator(self):
        """Check that the priority queue in the training simulator works as
        expected by creating a training simulator with a known list of
        event start times and event durations.
        """
        event_list = [
            EventTimingInfo(prev_event_start_to_current_start=1, duration=3),
            EventTimingInfo(prev_event_start_to_current_start=2, duration=5),
            EventTimingInfo(prev_event_start_to_current_start=2, duration=1),
            EventTimingInfo(prev_event_start_to_current_start=10, duration=10),
        ]
        start_times_gaps = [val.prev_event_start_to_current_start for val in event_list]
        start_times = [
            sum(start_times_gaps[0 : (x + 1)]) for x in range(0, len(start_times_gaps))
        ]
        durations = [d.duration for d in event_list]
        end_times = [t[0] + t[1] for t in zip(start_times, durations)]
        distr = AsyncTrainingEventGeneratorFromList(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorFromListConfig(training_events=event_list)
            )
        )
        # time-sorted event list:
        # event0 starts at time  1
        # event1 starts at time  3
        # event0 ends at time    4
        # event2 starts at time  5
        # event2 ends at time    6
        # event1 ends at time    8
        num_users = 50
        job_scheduler = TestAsyncTrainingEventHandler()
        data_provider = self._create_data_provider(
            num_users=num_users, examples_per_user=1
        )
        training_sim = AsyncTrainingSimulator(
            job_scheduler=job_scheduler,
            user_selector=self.random_user_selector(data_provider=data_provider),
            event_generator=distr,
            shared_client_config=self.shared_client_config,
            num_train_end_events_per_epoch=len(event_list) - 1,
        )
        training_sim.run_one_epoch()
        # events are stored in increasing order of ending time
        seq_index = 0
        event_index = 0
        # 0: event0 starts at time 1
        self._verify_event(
            job_scheduler.training_events[seq_index],
            start_times[event_index],
            end_times[event_index],
            TrainingState.TRAINING,
        )
        # 1: event1 starts at time 3
        seq_index += 1
        event_index = 1
        self._verify_event(
            job_scheduler.training_events[seq_index],
            start_times[event_index],
            end_times[event_index],
            TrainingState.TRAINING,
        )
        # 2: event0 ends at time 4
        seq_index += 1
        event_index = 0
        self._verify_event(
            job_scheduler.training_events[seq_index],
            start_times[event_index],
            end_times[event_index],
            TrainingState.TRAINING_FINISHED,
        )
        # 3: event2 starts time 5
        seq_index += 1
        event_index = 2
        self._verify_event(
            job_scheduler.training_events[seq_index],
            start_times[event_index],
            end_times[event_index],
            TrainingState.TRAINING,
        )
        # 4: event2 ends time 6
        seq_index += 1
        event_index = 2
        self._verify_event(
            job_scheduler.training_events[seq_index],
            start_times[event_index],
            end_times[event_index],
            TrainingState.TRAINING_FINISHED,
        )
        # 5: event1 ends time 8
        seq_index += 1
        event_index = 1
        self._verify_event(
            job_scheduler.training_events[seq_index],
            start_times[event_index],
            end_times[event_index],
            TrainingState.TRAINING_FINISHED,
        )

        assertEqual(job_scheduler.current_seqnum, 3)
        # check stats for seqnum
        # global model updates are as below: {user_seqnum, global_seqnum}
        # {1, 1} event1 ends
        # {2, 2} event3 ends
        # {1, 3} event2 ends
        # seqnum_diffs = [0, 0, 2], mean = 2/3, sd = sqrt(24/27)
        assertAlmostEqual(job_scheduler.seqnum_diff_mean(), 2 / 3)
        assertAlmostEqual(job_scheduler.seqnum_std(), math.sqrt(24 / 27))

    def test_async_stats(self):
        """Check that the priority JobQueueStats functionality in the training
        simulator works as expected by creating a training simulator with a
        known list of event start times and event durations.
        """
        event_list = [
            EventTimingInfo(prev_event_start_to_current_start=1, duration=3),
            EventTimingInfo(prev_event_start_to_current_start=2, duration=5),
            EventTimingInfo(prev_event_start_to_current_start=2, duration=1),
            EventTimingInfo(prev_event_start_to_current_start=10, duration=10),
        ]
        distr = AsyncTrainingEventGeneratorFromList(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorFromListConfig(training_events=event_list)
            )
        )
        # time-sorted event list:
        # event1 starts at time  1
        # event2_starts at time  3
        # event1 ends at time    4, 2 pending jobs
        # event3 starts at time  5
        # event3 ends at time    6, 2 pending job
        # event2 ends at time    8, 1 pending jobs
        num_users = 50
        data_provider = self._create_data_provider(
            num_users=num_users, examples_per_user=1
        )
        job_scheduler = TestAsyncTrainingEventHandler()
        training_sim = AsyncTrainingSimulator(
            job_scheduler=job_scheduler,
            user_selector=self.random_user_selector(data_provider=data_provider),
            event_generator=distr,
            shared_client_config=self.shared_client_config,
            num_train_end_events_per_epoch=len(event_list) - 1,
        )
        training_sim.run_one_epoch()
        avg_pending_jobs = training_sim.avg_pending_jobs()
        assertTrue(avg_pending_jobs == 5 / 3)

    def test_sequential_training(self) -> None:
        """Check that in sequential training (where mean and SD of training
        time is zero), jobs are truly trained sequentialy: i.e, if job A
        starts training, job A ends training before any other jobs start
        training
        """
        num_users = 50
        training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
            training_rate=10
        )
        duration_distr = PerExampleGaussianDurationDistributionConfig(
            training_duration_mean=0, training_duration_sd=0
        )
        distr = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        job_scheduler = TestAsyncTrainingEventHandler()
        data_provider = self._create_data_provider(
            num_users=num_users, examples_per_user=1
        )
        training_sim = AsyncTrainingSimulator(
            job_scheduler=job_scheduler,
            user_selector=self.random_user_selector(data_provider=data_provider),
            event_generator=distr,
            # pyre-ignore[16]: for pytest fixture
            shared_client_config=self.shared_client_config,
            num_train_end_events_per_epoch=num_users,
        )
        training_sim.run_one_epoch()
        # two 'events' are generated for each job,
        # and stored in job_scheduler.training_events
        # one TRAINING event and one TRAINING_FINISHED event
        assertEqual(len(job_scheduler.training_events), num_users * 2)
        for user_num in range(num_users):
            # two events for each user
            user_index = 2 * user_num
            assertEqual(
                job_scheduler.training_events[user_index][Globals.TRAINING_STATE_IDX],
                TrainingState.TRAINING,
            )
            assertEqual(
                job_scheduler.training_events[user_index + 1][
                    Globals.TRAINING_STATE_IDX
                ],
                TrainingState.TRAINING_FINISHED,
            )
            # verify that both events are from the same user
            assertEqual(
                job_scheduler.training_events[user_index][Globals.DEVICE_STATE_IDX],
                job_scheduler.training_events[user_index + 1][Globals.DEVICE_STATE_IDX],
            )

    def test_uniform_event_gen_stats(self) -> None:
        """Test that when using an EventGenerator with uniformly distributed
        time between events, and constant duration of training events,
        stats about #pending_jobs and model_seqnum_diff are correct
        """
        for test_config in [
            ConstantEventGenTestConfig,
            ConstantEventGenTestConfigPerUserGaussian,
        ]:
            num_users = test_config.num_users
            num_examples_per_user = test_config.num_examples_per_user
            training_start_time_distr = ConstantAsyncTrainingStartTimeDistrConfig(
                training_rate=test_config.training_rate
            )
            training_duration_mean = test_config.training_duration_mean
            training_duration_sd = test_config.training_duration_sd
            duration_distr = test_config.training_duration_distribution_config(
                training_duration_mean=training_duration_mean,
                training_duration_sd=training_duration_sd,
            )
            distr = AsyncTrainingEventGenerator(
                **OmegaConf.structured(
                    AsyncTrainingEventGeneratorConfig(
                        training_start_time_distribution=training_start_time_distr,
                        duration_distribution_generator=duration_distr,
                    )
                )
            )
            job_scheduler = TestAsyncTrainingEventHandler()
            data_provider = self._create_data_provider(
                num_users=num_users, examples_per_user=num_examples_per_user
            )
            training_sim = AsyncTrainingSimulator(
                job_scheduler=job_scheduler,
                user_selector=self.random_user_selector(data_provider=data_provider),
                event_generator=distr,
                # pyre-ignore[16]: for pytest fixture
                shared_client_config=self.shared_client_config,
                num_train_end_events_per_epoch=num_users,
            )
            training_sim.run_one_epoch()
            assertEqual(
                training_sim.queue_stats.avg_pending_jobs(),
                test_config.mean_pending_jobs,
            )
            assertEqual(job_scheduler.seqnum_diff_mean(), test_config.mean_seqnum_diff)
            assertEqual(job_scheduler.seqnum_std(), test_config.sd_seqnum_diff)

    def test_poisson_event_gen_stats(self) -> None:
        """Test that when using an EventGenerator with poisson distributed
        time between events, and constant duration of training events,
        stats about #pending_jobs and model_seqnum_diff are correct
        """
        for test_config in [
            PoissonEventGenTestConfig,
            PoissonEventGenTestConfigPerUserGaussian,
        ]:
            np.random.seed(1)
            torch.manual_seed(1)
            num_users = test_config.num_users
            num_examples_per_user = test_config.num_examples_per_user
            training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
                training_rate=test_config.training_rate
            )
            duration_distr = PerExampleGaussianDurationDistributionConfig(
                training_duration_mean=test_config.training_duration_mean,
                training_duration_sd=test_config.training_duration_sd,
            )
            distr = AsyncTrainingEventGenerator(
                **OmegaConf.structured(
                    AsyncTrainingEventGeneratorConfig(
                        training_start_time_distribution=training_start_time_distr,
                        duration_distribution_generator=duration_distr,
                    )
                )
            )
            job_scheduler = TestAsyncTrainingEventHandler()
            data_provider = self._create_data_provider(
                num_users=num_users, examples_per_user=num_examples_per_user
            )
            training_sim = AsyncTrainingSimulator(
                job_scheduler=job_scheduler,
                user_selector=self.random_user_selector(data_provider=data_provider),
                event_generator=distr,
                # pyre-ignore[16]: for pytest fixture
                shared_client_config=self.shared_client_config,
                num_train_end_events_per_epoch=num_users,
            )
            training_sim.run_one_epoch()
            assertAlmostEqual(
                training_sim.queue_stats.avg_pending_jobs(),
                test_config.mean_pending_jobs,
                delta=0.50,
            )
            assertEqual(job_scheduler.seqnum_diff_mean(), test_config.mean_seqnum_diff)
            assertAlmostEqual(
                job_scheduler.seqnum_std(), test_config.sd_seqnum_diff, delta=0.001
            )

    @staticmethod
    def get_user_finishing_seq(
        total_num_users: int,
        training_duration_distr_config: Type[DurationDistributionConfig],
    ) -> List[int]:
        per_user = PerUserGaussianDurationDistributionConfig
        per_example = PerExampleGaussianDurationDistributionConfig
        if training_duration_distr_config == per_user:
            # sequential finishing order
            return list(range(total_num_users))
        elif training_duration_distr_config == per_example:
            # reverse finishing order
            return list(reversed(range(total_num_users)))
        else:
            raise AssertionError(
                f"Unknown type string:{training_duration_distr_config}"
            )

    def verify_num_examples_multiplier(
        self, duration_distr_config: Type[DurationDistributionConfig]
    ) -> None:
        """if duration_distr_type == per-example-gaussian
           training_duration=num_examples*training_duration_per_example
        else
           training_duration=training_duration_per_example

        Use a config where training time is completely determined by
        number of examples. Eg:
            - GaussianDuration,PoissonStartTime
            - training_rate: very high
            - mean training time: very high
        num_examples_per_user = [n, n-1, n-2, ..., 3, 2, 1]

        If per-example-gaussian
            The last user should finish first.
            So training finish time should be: [user n, user n-1, ..., user2, user1]
        else if per-user-gaussian
            First user should finish first
            So training finish time should be: [user 1, user 2, user 3, .... user n]
        """
        num_users = 50
        training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
            training_rate=1000
        )
        duration_distr = duration_distr_config(
            training_duration_mean=1000,
            training_duration_sd=0,
        )
        distr = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        num_examples_per_user = list(reversed(range(1, num_users + 1)))
        data = [
            [1] * num_example
            for num_example, _ in zip(num_examples_per_user, range(num_users))
        ]
        data_provider = FLDataProviderFromList(
            train_user_list=data,
            eval_user_list=data,
            test_user_list=data,
            model=MockFLModel(),
        )

        job_scheduler = TestAsyncTrainingEventHandler()
        training_sim = AsyncTrainingSimulator(
            job_scheduler=job_scheduler,
            user_selector=self.round_robin_user_selector(data_provider),
            event_generator=distr,
            # pyre-ignore[16]: for pytest fixture
            shared_client_config=self.shared_client_config,
            num_train_end_events_per_epoch=num_users,
        )
        training_sim.run_one_epoch()
        ts_idx = Globals.TRAINING_STATE_IDX
        ds_idx = Globals.DEVICE_STATE_IDX
        # two 'events' are generated for each job,
        # and stored in job_scheduler.training_events
        # one TRAINING event and one TRAINING_FINISHED event
        assertEqual(len(job_scheduler.training_events), num_users * 2)
        first_n_events = job_scheduler.training_events[:num_users]
        # First num_user events should be TRAINING events
        assertEqual(
            [i[ts_idx] for i in first_n_events], [TrainingState.TRAINING] * num_users
        )
        # First num_user events should have user_index in sequence
        assertEqual(
            [i[ds_idx].user_info.user_index for i in first_n_events],
            list(range(num_users)),
        )

        last_n_events = job_scheduler.training_events[num_users:]
        # Next num_user events should be TRAINING_FINISHED events:
        assertEqual(
            [i[ts_idx] for i in last_n_events],
            [TrainingState.TRAINING_FINISHED] * num_users,
        )
        # Users should finish in seq if type = PerUserGaussian,
        # in reverse seq if type = PerExampleGaussian
        user_finishing_seq = self.get_user_finishing_seq(
            num_users, duration_distr_config
        )
        assertEqual(
            [i[ds_idx].user_info.user_index for i in last_n_events], user_finishing_seq
        )

    def test_training_duration_per_example_gaussian(self):
        # TODO: T88456330: Fix timing out issue for this test
        pass

    def test_training_duration_per_user_gaussian(self):
        # TODO: T88456330: Fix timing out issue for this test
        pass

    def test_training_end_training_start_relative_priority(self):
        r"""
        Test that when training end and training start events have the same time,
        training end event happens earlier (has lower priority)
        """

    def test_logging_level(self):
        num_users = 50
        training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
            training_rate=10
        )
        duration_distr = PerExampleGaussianDurationDistributionConfig(
            training_duration_mean=0, training_duration_sd=0
        )
        distr = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        job_scheduler = TestAsyncTrainingEventHandler()
        data_provider = self._create_data_provider(
            num_users=num_users, examples_per_user=1
        )
        training_sim = AsyncTrainingSimulator(
            job_scheduler=job_scheduler,
            user_selector=self.random_user_selector(data_provider=data_provider),
            event_generator=distr,
            shared_client_config=self.shared_client_config,
            num_train_end_events_per_epoch=num_users,
        )
        assertTrue(check_inherit_logging_level(training_sim, 50))
        assertTrue(check_inherit_logging_level(training_sim, 10))

    def _training_staleness_distribution(
        self,
        duration_distr,
        training_rate=1000,
        num_users=10000,
        examples_per_user=1,
    ):
        data_provider = self._create_data_provider(
            num_users=num_users, examples_per_user=examples_per_user
        )
        training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
            training_rate=training_rate
        )
        distr = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        handler = TestAsyncTrainingEventHandler()

        training_sim = AsyncTrainingSimulator(
            job_scheduler=handler,
            user_selector=RandomAsyncUserSelector(data_provider=data_provider),
            event_generator=distr,
            shared_client_config=self.shared_client_config,
            num_train_end_events_per_epoch=num_users,
        )
        training_sim.run_one_epoch()
        return np.array(handler.num_unseen_global_model_updates)

    def test_per_user_half_normal_training_duration(self):
        """
        Test for staleness with half normal

        Condition 1:
            Check right skewed to make sure that the distribution is
            heavy tailed and there is no concentration of users with low
            staleness more than usual
        Condition 2:
            Check that the distribution has a long tail that approximate a guassian tail
            All values should be within 6 std of the mean
        """
        num_users = 1000
        training_rate = 10
        training_min = 0.00
        training_std = 1.5
        duration_distr_config = PerUserHalfNormalDurationDistributionConfig(
            training_duration_sd=training_std, training_duration_min=training_min
        )
        staleness = self._training_staleness_distribution(
            duration_distr_config, training_rate=training_rate, num_users=num_users
        )
        mean = np.mean(staleness)
        sd = np.std(staleness)

        assertLessEqual(np.median(staleness), mean)
        # check that not all values are within 2 sd
        assertGreaterEqual(max(staleness), mean + 2 * sd)
        # check all values are within 7 sd of the mean
        # 2e-11 = e^(−k^2/2) = exp(-7^2/2)
        assertLessEqual(max(staleness), mean + 7 * sd)

    def test_per_user_uniform_training_duration(self):
        num_users = 2000
        training_rate = 10
        training_min = 0.0
        training_mean = 1.0

        duration_distr_config = PerUserUniformDurationDistributionConfig(
            training_duration_mean=training_mean, training_duration_min=training_min
        )
        staleness = self._training_staleness_distribution(
            duration_distr_config, training_rate=training_rate, num_users=num_users
        )
        mean = np.mean(staleness)
        sd = np.std(staleness)

        # check that under uniform mean and median close to the same
        assertAlmostEqual(np.median(staleness), mean, delta=1.5)

        # check that all values are within 4 sd from the mean
        assertLessEqual(max(staleness), mean + 4 * sd)
        assertGreaterEqual(min(staleness), mean - 4 * sd)

    def test_per_user_exponential_training_duration(self):
        for training_mean in [0.5, 1, 2]:
            num_users = 2000
            training_rate = 10

            duration_distr_config = PerUserExponentialDurationDistributionConfig(
                training_duration_mean=training_mean
            )
            staleness = self._training_staleness_distribution(
                duration_distr_config, training_rate=training_rate, num_users=num_users
            )
            mean = np.mean(staleness)
            median = np.median(staleness)

            expected_mean = training_rate * training_mean
            assertAlmostEqual(expected_mean, mean, delta=1.5)
            assertLessEqual(median, mean)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from dataclasses import dataclass
from enum import Enum, auto

from flsim.utils.async_trainer.training_event_generator import IEventGenerator


class TrainingState(Enum):
    # Orderinig is important
    # For devices that have the same next_event_time(), we want devices that
    # "further along" in training to be chosen first
    # hence, TRAINING_FINISHED < TRAINING < WAITING_FOR_START
    TRAINING_FINISHED = auto()
    TRAINING = auto()
    WAITING_FOR_START = auto()

    # https://docs.python.org/3/library/enum.html#orderedenum
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


@dataclass
class TrainingSchedule:
    r"""
    Class to represent a client training time duration
    """
    creation_time: float
    start_time: float
    end_time: float


class TrainingScheduleFactory:
    @classmethod
    def create(
        cls,
        current_time: float,
        event_generator: IEventGenerator,
        num_examples: int,
    ):
        creation_time = current_time
        start_time = creation_time + event_generator.time_to_next_event_start()
        duration = event_generator.training_duration(num_examples)
        end_time = start_time + duration
        return TrainingSchedule(creation_time, start_time, end_time)


class DeviceState:
    r"""
    Represents the state of a device that's either waiting to start training,
    or in the middle of training
    """

    def __init__(self, training_schedule: TrainingSchedule):
        self.training_schedule: TrainingSchedule = training_schedule
        self.training_state: TrainingState = TrainingState.WAITING_FOR_START

    def get_training_state(self):
        return self.training_state

    # when we start training, we get initial model as input
    def training_started(self) -> None:
        self.training_state = TrainingState.TRAINING

    def training_ended(self) -> None:
        self.training_state = TrainingState.TRAINING_FINISHED

    def next_event_time(self):
        if self.training_state == TrainingState.WAITING_FOR_START:
            return self.training_schedule.start_time
        else:
            return self.training_schedule.end_time

    def __lt__(self, other):
        # if two device states have the same 'next_event_time', chose the one
        # that has 'smaller' training_state
        # smaller training state => further along in training
        if self.next_event_time() == other.next_event_time():
            return self.training_state < other.training_state
        return self.next_event_time() < other.next_event_time()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import torch
from flsim.common.pytest_helper import assertEqual, assertTrue
from flsim.optimizers.optimizer_scheduler import (
    ArmijoLineSearch,
    ArmijoLineSearchSchedulerConfig,
)
from flsim.tests.utils import MockQuadratic1DFL, Quadratic1D
from omegaconf import OmegaConf


class TestOptimizerScheduler:
    def test_armijo_line_search_on_parabola(self):
        """
            a toy optimization example:
                min f(x) = 100 x^2 - 1

        minima is x=0.0, x is initialized at 1.0.
        Gradient descent with constant step-size 0.01 will never
        converge, and in fact, jump between -1 and +1 interleavingly.
        In contrast, Armijo line-search reduces step-sizes to avoid
        "over-shooting" and converges to 0.
        """
        # set up quadratic parabola objective and optimizer
        quadratic1D = MockQuadratic1DFL(Quadratic1D())
        optimizer = torch.optim.SGD(
            quadratic1D.fl_get_module().parameters(), lr=0.01, momentum=0.0
        )
        # run (deterministic) GD for 10 steps with step-size = 0.01,
        # with constant step-size = 0.01, and x0 = 1.0, the iteration
        # never converge and will jump between -1.0 and 1.0 interleavingly
        for i in range(10):
            optimizer.zero_grad()
            metrics = quadratic1D.fl_forward()
            quadratic_func_val = metrics.loss
            quadratic_func_val.backward()
            optimizer.step()
            obj_val = quadratic1D.fl_get_module().state_dict()["x"].item()
            assertEqual(obj_val, (-1.0) ** (i + 1))

        # set up (again) quadratic parabola objective and optimizer
        quadratic1D = MockQuadratic1DFL(Quadratic1D())
        optimizer = torch.optim.SGD(
            quadratic1D.fl_get_module().parameters(), lr=0.01, momentum=0.0
        )
        # use Armijo line-search for optimizer step-size selection
        # same initial step-size
        config = ArmijoLineSearchSchedulerConfig()
        config.base_lr = 0.01
        config.reset = True
        armijo_line_search_scheduler = ArmijoLineSearch(
            optimizer=optimizer, **OmegaConf.structured(config)
        )
        # run for 10 steps
        for t in range(10):
            optimizer.zero_grad()
            metrics = quadratic1D.fl_forward()
            quadratic_func_val = metrics.loss
            quadratic_func_val.backward()
            armijo_line_search_scheduler.step(metrics, quadratic1D, None, t)
            optimizer.step()
        # check converging to 0 (true answer)
        assertTrue(quadratic1D.fl_get_module().state_dict()["x"].item() <= 1e-7)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass

import torch
from flsim.channels.base_channel import (
    IdentityChannel,
    FLChannelConfig,
    Message,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg


class SparseMaskChannel(IdentityChannel):
    """
    Implements a channel where the message sent from client to server is
    masked by the top-k absolute values in the model parameter weights.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=SparseMaskChannelConfig,
            **kwargs,
        )
        super().__init__()
        self.proportion_of_zero_weights = self.cfg.proportion_of_zero_weights
        self.sparsity_method = self.cfg.sparsity_method
        self.compressed_size_measurement = self.cfg.compressed_size_measurement
        assert self.cfg.sparsity_method in {
            "random",
            "topk",
        }, "Compression method must be one of 'random' or 'topk'"
        assert (
            0 <= self.cfg.proportion_of_zero_weights < 1
        ), "Compression rate must be in [0, 1)"
        assert self.cfg.compressed_size_measurement in {
            "coo",
            "bitmask",
        }, "Compressed size measurement must be one of 'coo' or 'bitmask'"

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _calc_message_size_client_to_server(self, message: Message):
        """
        For COO format : We compute the size of the compressed message as follows: for any
        parameter, we count the number of non-zero entries. Then, we assume
        that the sparse tensor is stored in the COO format: each non-zero
        entry is stored with a value (in fp32) and an index (int64 * ndim).

        For bitmask format : We compute the size of the compressed message as follows when
        serial format is specified: the non-zero entries are transmitted as is in fp32 format,
        and the sparsity mask is transmitted as a stream of bits to represent sparse locations.
        """
        message_size_bytes = 0
        for param in message.model_state_dict.values():

            # get number of non-sparse entries (nse)
            nse = param.numel() - int(self.proportion_of_zero_weights * param.numel())
            # size of the index
            if self.compressed_size_measurement == "coo":
                message_size_bytes += (
                    param.ndim * SparseMaskChannel.BYTES_PER_INT64 * nse
                )
            elif self.compressed_size_measurement == "bitmask":
                message_size_bytes += param.numel() * SparseMaskChannel.BYTES_PER_BIT

            # size of the values
            message_size_bytes += nse * SparseMaskChannel.BYTES_PER_FP32
        return message_size_bytes

    def _on_server_before_transmission(self, message: Message) -> Message:
        message.populate_state_dict()
        return message

    def _on_client_before_transmission(self, message: Message) -> Message:
        """
        Here we apply a sparse mask to the parameter updates before sending the message.

        Notes:
            - There are two options for sparsity: random and topk
            - In random sparsity, the mask is randomly selecting parameter weight updates
            - In TopK sparsity, sparsity is applied on each parameter's weight update
              separately depending on the magnitude of the values; the smallest values
              get pruned.
            - The message is pruned so that the number of non-sparse entries is
              deterministc and constant across runs for a given weight matrix.
        """
        message.populate_state_dict()
        new_state_dict = OrderedDict()

        for name, param in message.model_state_dict.items():
            # exact number of elements to prune
            num_params_to_prune = int(self.proportion_of_zero_weights * param.numel())

            # select flat indices to prune
            top_k = torch.topk(
                (
                    torch.rand_like(param.data)
                    if self.sparsity_method == "random"
                    else torch.abs(param.data)
                ).view(-1),
                k=num_params_to_prune,
                largest=False,
            )

            # prune top-K
            param.data.view(-1)[top_k.indices] = 0
            new_state_dict[name] = param.data

        message.model_state_dict = new_state_dict
        return message


@dataclass
class SparseMaskChannelConfig(FLChannelConfig):
    _target_: str = fullclassname(SparseMaskChannel)
    proportion_of_zero_weights: float = 0.5
    sparsity_method: str = "random"
    compressed_size_measurement: str = "bitmask"



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from typing import Type

import pytest
import torch
from flsim.channels.communication_stats import ChannelDirection
from flsim.channels.message import Message
from flsim.channels.sketch_channel import (
    SketchChannelConfig,
    SketchChannel,
)
from flsim.common.pytest_helper import (
    assertAlmostEqual,
    assertEqual,
    assertIsInstance,
    assertRaises,
    assertTrue,
)
from flsim.tests import utils
from flsim.utils.count_sketch import CountSketch
from hydra.utils import instantiate


class TestCountSketch:
    def test_count_sketch_creation(self) -> None:
        """
        Tests instantiation of the count sketch channel.
        """

        good_hash = torch.randint(0, 10, (10, 2))
        bad_hash = torch.randint(0, 10, (2, 1))

        cs = CountSketch(width=10, depth=10, h=good_hash, g=good_hash)
        assertIsInstance(cs, CountSketch)

        with assertRaises(AssertionError):
            _ = CountSketch(width=10, depth=5, h=bad_hash)

        with assertRaises(AssertionError):
            _ = CountSketch(width=10, depth=5, g=bad_hash)

    def test_count_sketch_update_query(self) -> None:
        """
        Tests if it is possible to sketch and unsketch one number using CountSketch.
        Also check that unsketching an id that were never sketched would return 0.
        """

        cs = CountSketch(width=10001, depth=11, prime=2 ** 31 - 1, independence=4)
        # test simple update and query
        cs.update(torch.tensor([0]), torch.tensor([2.0]))
        assertTrue(torch.isclose(cs.query(torch.tensor([0])), torch.tensor((2.0))))
        assertTrue(torch.isclose(cs.query(torch.tensor([1])), torch.tensor((0.0))))
        assertEqual(torch.sum(torch.abs(cs.buckets)).item(), 2 * 11)

    def test_count_sketch_model(self) -> None:
        """
        Tests if CountSketch stores the correct parameter names and tensor sizes
        after sketching a model and the unsketched model has approximately the same
        weights.
        """

        cs = CountSketch(width=10001, depth=11, prime=2 ** 31 - 1, independence=4)
        two_fc = utils.TwoFC()
        two_fc.fill_all(0.2)
        # should have 21 parameters
        model = utils.SampleNet(two_fc)

        cs.sketch_model(model=model)

        for (cs_name, cs_param), (model_name, model_param) in zip(
            list(cs.param_sizes.items()), list(model.fl_get_module().named_parameters())
        ):
            assertEqual(cs_name, model_name)
            assertEqual(cs_param, model_param.size())

        for (_, cs_param) in cs.unsketch_model().items():
            assertTrue(
                False
                not in torch.isclose(
                    cs_param, torch.full_like(cs_param, fill_value=0.2)
                )
            )

    def test_count_sketch_size(self) -> None:
        width = 1000
        depth = 10
        cs = CountSketch(width=width, depth=depth)
        size = cs.get_size_in_bytes()
        assertAlmostEqual(size, width * depth * 4, delta=1)


class TestSketchChannelTest:
    @pytest.mark.parametrize(
        "config",
        [SketchChannelConfig(num_col=1000, num_hash=7, prime=2 ** 13 - 1)],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [SketchChannel],
    )
    def test_sketch_channel(self, config: Type, expected_type: Type) -> None:
        """
        Tests client to server direction. Models should be almost
        identical since we use a very large sketch wrt model size.
        """

        channel = instantiate(config)
        assertIsInstance(channel, expected_type)

        two_fc = utils.TwoFC()
        two_fc.fill_all(0.2)
        model = utils.SampleNet(two_fc)

        message = Message(model)
        message = channel.client_to_server(message)
        cs = message.count_sketch
        assertIsInstance(cs, CountSketch)

        # same test as the count_sketch test
        for (cs_name, cs_param), (model_name, model_param) in zip(
            list(cs.param_sizes.items()), list(model.fl_get_module().named_parameters())
        ):
            assertEqual(cs_name, model_name)
            assertEqual(cs_param, model_param.size())

        for (_, cs_param) in cs.unsketch_model().items():
            assertTrue(
                False
                not in torch.isclose(
                    cs_param, torch.full_like(cs_param, fill_value=0.2)
                )
            )

    @pytest.mark.parametrize(
        "config",
        [
            SketchChannelConfig(
                num_col=1000,
                num_hash=7,
                prime=2 ** 13 - 1,
                report_communication_metrics=True,
            )
        ],
    )
    def test_sketch_stats(self, config: Type) -> None:
        """
        Tests stats measurement. We manually compute bytes sent from client
        to the server and verify that it matches the bytes computed by the
        channel.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # client -> server
        message = Message(upload_model)
        message = channel.client_to_server(message)

        # test communication stats measurements
        stats = channel.stats_collector.get_channel_stats()
        client_to_server_bytes = stats[ChannelDirection.CLIENT_TO_SERVER].mean()

        # the sketch has `num_col` elements per hash. Assuming fp32 elements.
        computed_sketch_size_bytes = (
            config.num_col * config.num_hash * channel.BYTES_PER_FP32
        )

        assertEqual(client_to_server_bytes, computed_sketch_size_bytes)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from typing import Any, Dict

# see https://fb.workplace.com/groups/fbcode/permalink/2440457449324413/
# @manual=third-party//pandas:pandas-py
import pandas as pd
from flsim.interfaces.dataset import FLDataset


class FLCSVDataset(FLDataset):
    """Abstract class that can be easily extended by implementing one light
    method (_get_processed_row_from_single_raw_row()) that is based on
    PyTorch's Dataset class. For example,
    ```
    class AdsCVRDataset(FLCSVDataset):
        def _get_processed_row_from_single_raw_row(self, raw_row: Any):
            return {
                "float_features": torch.Tensor(featurize(raw_row["feat"])),
                "label": torch.Tensor(raw_row["label"]),
            }
    ```
    then, one can simply enumerate over this Dataset as s/he usually doee
    with PyTorch Dataset to iterate over training/eval/test batches.
    However, note that you will most likely be encapsulating this within
    a concrete implementation of IFLDataLoader to use this within FL
    simulation with minimal effort (e.g. see FLAdsDataLoader).
    """

    def __init__(self, path: str):
        self.data_frame = pd.read_csv(path)

    def __len__(self):
        return len(self.data_frame)

    def __getitem__(self, idx):
        raw_row = self.data_frame.iloc[idx]
        return self._get_processed_row_from_single_raw_row(raw_row)

    @abc.abstractmethod
    def _get_processed_row_from_single_raw_row(self, raw_row: Any) -> Dict[str, Any]:
        """This method should be overridden by the child class. One should
        provide logic to convert a raw row read from data into a processed
        row (i.e. dictionary where key is column name and value is whatever
        model expects such as tensor for features and integer for labels).
        """
        pass



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Any, Dict

import pkg_resources
import pytest
import torch
from flsim.common.pytest_helper import assertEqual
from flsim.data.csv_dataset import FLCSVDataset
from flsim.data.data_sharder import ColumnSharderConfig
from flsim.data.dataset_data_loader import FLDatasetDataLoaderWithBatch
from hydra.utils import instantiate


@pytest.fixture(scope="class")
def prepare_dataset_data_loader_with_batch(request):
    request.cls.test_csv_path = "test_resources/data.csv"
    request.cls.total_data_count = 15
    request.cls.train_batch_size = 1
    request.cls.eval_batch_size = 3
    request.cls.test_batch_size = 5


class TestDataset(FLCSVDataset):
    def _get_processed_row_from_single_raw_row(self, raw_row: Any) -> Dict[str, Any]:
        return {
            "userid": torch.Tensor([raw_row["userid"]]),
            "label": torch.Tensor([raw_row["label"]]),
        }


@pytest.mark.usefixtures("prepare_dataset_data_loader_with_batch")
class TestDatasetDataLoaderWithBatch:
    def test_batch_size(self) -> None:
        # pyre-ignore[16]: for pytest fixture
        file_path = pkg_resources.resource_filename(__name__, self.test_csv_path)
        dataset = TestDataset(file_path)

        fl_data_sharder = instantiate(ColumnSharderConfig(sharding_col="userid"))
        data_loader = FLDatasetDataLoaderWithBatch(
            dataset,
            dataset,
            dataset,
            fl_data_sharder,
            # pyre-ignore[16]: for pytest fixture
            self.train_batch_size,
            # pyre-ignore[16]: for pytest fixture
            self.eval_batch_size,
            # pyre-ignore[6]
            self.test_batch_size,
        )
        assertEqual(
            len(list(data_loader.fl_train_set())),
            # pyre-ignore[16]: for pytest fixture
            self.total_data_count / self.train_batch_size,
        )
        assertEqual(
            len(list(data_loader.fl_eval_set())),
            self.total_data_count / self.eval_batch_size,
        )
        assertEqual(
            len(list(data_loader.fl_test_set())),
            self.total_data_count / self.test_batch_size,
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import argparse
import collections
import json
from typing import Any, Dict, List, Tuple, Type

from hydra.experimental import initialize, compose
from omegaconf import OmegaConf, DictConfig


def fullclassname(cls: Type[Any]) -> str:
    """
    Returns the fully qualified class name of the input class.
    """
    module = cls.__module__
    name = cls.__qualname__
    if module is not None and module != "__builtin__":
        name = module + "." + name
    return name


def _validate_cfg(component_class: Type[Any], cfg: Any):
    """
    Validate that cfg doesn't have MISSING fields. This needs to be done only after
    all defaults are set, typically in the base class.
    We do this by making sure none of the parents have ``_set_defaults_in_cfg`` method.
    """
    if not any(
        hasattr(parent, "_set_defaults_in_cfg") for parent in component_class.__bases__
    ):
        # looping over the config fields throws incase of missing field
        for _ in cfg.items():
            pass


def init_self_cfg(
    component_obj: Any,
    *,
    component_class: Type,
    config_class: Type,
    **kwargs,
):
    """
    Initialize FL component config by constructing OmegaConf object,
    setting defaults, and validating config.
    """
    cfg = (
        config_class(**kwargs)
        if not hasattr(component_obj, "cfg")
        else component_obj.cfg
    )
    cfg = OmegaConf.create(cfg)  # convert any structure to OmegaConf
    component_class._set_defaults_in_cfg(cfg)  # set default cfg params for this class
    # convert any structure to OmegaConf again, after setting defaults
    cfg = OmegaConf.create(cfg)  # pyre-ignore [6]
    _validate_cfg(component_class, cfg)  # validate the config
    component_obj.cfg = cfg


# trainer config utils for consuming hydra configs
def _flatten_dict(
    d: collections.MutableMapping, parent_key="", sep="."
) -> Dict[str, str]:
    """
    Changes json of style
    ```
    {
        "trainer" : {
            "_base_": "base_sync_trainer",
            "aggregator": {
                "_base_": "base_fed_avg_with_lr_sync_aggregator",
                "lr": 0.1
            }
        }
    }
    ```
    to
    ```
    {
        "trainer._base_": "base_sync_trainer",
        "trainer.aggregator._base_": "base_fed_avg_with_lr_sync_aggregator",
        "trainer.aggregator.lr": 0.1,
    }
    ```
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        # if value is not a dict and is mutable, extend the items and flatten again.
        # > hacky way of preserving dict values by checking if key has _dict as suffix.
        if not new_key.endswith("_dict") and isinstance(v, collections.MutableMapping):
            items.extend(_flatten_dict(v, new_key, sep=sep).items())
        else:
            # check if a number needs to be retained as a string
            # the repalce with one dot is needed to handle floats
            if type(v) is str and v.replace(".", "", 1).isdigit():
                v = f'"{v}"'  # enclose it with quotes if so.
            items.append((new_key, v))
    return dict(items)


def _handle_values_for_overrides_list(v: Any) -> Any:
    """
    Handle the special massaging of some values of JSON need to for it to be supplied
    to Hydra's overrides list.
    """
    # python's None --> cmd line null for override list
    v = "null" if v is None else v

    # if value is a dict, convert it to string to work with override list.
    # dump twice to escape quotes correctly.
    v = json.dumps(json.dumps(v)) if type(v) is dict else v
    # escape = char in value when present
    v = v.replace(r"=", r"\=") if type(v) is str else v
    return v


def _hydra_merge_order(dotlist_entry: str) -> Tuple:
    """
    The override list needs to be ordered as the last one wins in case of
    duplicates: https://hydra.cc/docs/advanced/defaults_list#composition-order
    This function arranges the list so that _base_ is at the top, and we
    proceed with overrides from top to bottom.
    """
    key = dotlist_entry.split("=")[0]
    # presence of "@" => it is a _base_ override
    default_list_item_indicator = key.count("@")  # 1 if true, 0 otherwise
    # level in hierarchy; based on number of "."
    hierarchy_level = key.count(".")
    # multiply by -1 to keep the default list items on top
    return (-1 * default_list_item_indicator, hierarchy_level, dotlist_entry)


def fl_json_to_dotlist(
    json_config: Dict[str, Any], append_or_override: bool = True
) -> List[str]:
    """
    Changes
    ```
    {
        "trainer._base_": "base_sync_trainer",
        "trainer.aggregator._base_": "base_fed_avg_with_lr_sync_aggregator",
        "trainer.aggregator.lr": 0.1,
    }
    ```
    to
    ```
    [
        "+trainer@trainer=base_sync_trainer",
        "+aggregator@trainer.aggregator=base_fed_avg_with_lr_sync_aggregator",
        "trainer.aggregator.lr=0.1",
    ]
    ```
    The override list grammar for reference:
        https://hydra.cc/docs/advanced/override_grammar/basic
    """
    dotlist_dict = _flatten_dict(json_config)
    dotlist_list = []
    for k, v in dotlist_dict.items():
        if k.endswith("._base_"):
            # trainer.aggregator._base_ --> trainer.aggregator
            k = k.replace("._base_", "")
            # extract aggregator from trainer.aggregator
            config_group = k.split(".")[-1]
            # trainer.aggregator --> +aggregator@trainer.aggregator
            k = f"+{config_group}@{k}"
            # +aggregator@trainer.aggregator=base_fed_avg_with_lr_sync_aggregator
            dotlist_list.append(f"{k}={v}")
        else:
            v = _handle_values_for_overrides_list(v)
            prefix = "++" if append_or_override else ""
            dotlist_list.append(f"{prefix}{k}={v}")
    sorted_dotlist_list = sorted(dotlist_list, key=_hydra_merge_order)
    return sorted_dotlist_list


def fl_config_from_json(
    json_config: Dict[str, Any], append_or_override: bool = True
) -> DictConfig:
    """
    Accepts the FLSim config in json format and constructs a Hydra config object.
    """
    with initialize(config_path=None):
        cfg = compose(
            config_name=None,
            overrides=fl_json_to_dotlist(json_config, append_or_override),
        )
        return cfg


def maybe_parse_json_config():
    """
    Parse the command line args and build a config object if json config is supplied.
    This comes in handy when we want to supply a json config file during to buck run.
    This function will no longer be relevant once FLSim entirely moves to YAML configs.
    """
    cfg = None
    parser = argparse.ArgumentParser(description="Run training loop for FL example")
    parser.add_argument("--config-file", type=str, default=None, help="JSON config")
    args, _ = parser.parse_known_args()
    # if JSON config is specified, build a DictConfig
    if args.config_file is not None:
        with open(args.config_file, "r") as config_file:
            json_config = json.load(config_file)
            cfg = fl_config_from_json(json_config["config"])
    # else:  assume yaml config, and let hydra handle config construction
    return cfg


def is_target(config, cls):
    return config._target_ == cls._target_



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from typing import List, Tuple

import numpy as np
from flsim.clients.async_client import AsyncClientDevice
from flsim.utils.async_trainer.device_state import DeviceState, TrainingState


class IAsyncTrainingEventHandler(abc.ABC):
    r"""
    Interface class that performs asynchronous FL model training.

    Handles start training for a particular device state and end training for a particular device state.
    The AsyncTrainer should implement this class along with FLTrainerBase
    """

    @abc.abstractmethod
    def on_training_start(self, client: AsyncClientDevice) -> None:
        r"""
        Callback to notify that a client is starting training

        Marks the client state as training.

        Note:
        -----
        The actual training process doesn't start until on_training_end

        There's a choice of when should training ACTUALLY happen: when the
        training_start event fires, or when training_end event fires
        Answer: when training_end event fires. That's because training often
        produces side effects - like metric collection. So we should only
        train when we're sure that we want those side-effects
        """
        pass

    @abc.abstractmethod
    def on_training_end(self, client: AsyncClientDevice) -> None:
        r"""
        Callback to notify the trainer of a training end event.

        Trains local model for the client and incoporates it to the global model
        """
        pass

    @abc.abstractmethod
    def train_and_update_global_model(self, client: AsyncClientDevice) -> None:
        pass


class AsyncTrainingEventHandler(IAsyncTrainingEventHandler):
    def on_training_start(self, client: AsyncClientDevice) -> None:
        pass

    def on_training_end(self, client: AsyncClientDevice) -> None:
        r"""
        Trains the client's local model and update seqnum

        This function does the following:
            1. Changes training state to `TRAINING_FINISHED`
            2. Trains client's local model by calling `train_and_update_global_model`
        """
        client.training_ended()
        self.train_and_update_global_model(client=client)

    def train_and_update_global_model(self, client: AsyncClientDevice) -> None:
        pass


class TestAsyncTrainingEventHandler(AsyncTrainingEventHandler):
    r"""
    AsyncTrainingEventHandler that stores all model_states in a list.
    Useful for testing training simulation.
    """

    def __init__(self):
        super().__init__()
        self.training_events: List[Tuple(DeviceState, TrainingState)] = []
        self.num_unseen_global_model_updates: List[int] = []
        self.current_seqnum = 0

    def on_training_start(self, client: AsyncClientDevice) -> None:
        client.training_started(self.current_seqnum)
        self.training_events.append((client, TrainingState.TRAINING))

    def on_training_end(self, client: AsyncClientDevice) -> None:
        super().on_training_end(client)
        self.training_events.append((client, TrainingState.TRAINING_FINISHED))
        self.current_seqnum += 1

    def train_and_update_global_model(self, client: AsyncClientDevice) -> None:
        self.num_unseen_global_model_updates.append(
            self.current_seqnum - client.model_seqnum
        )

    def seqnum_diff_mean(self):
        return np.mean(self.num_unseen_global_model_updates)

    def seqnum_std(self):
        return np.std(self.num_unseen_global_model_updates)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import hydra
import numpy as np
import pytest
from flsim.common.pytest_helper import assertEqual, assertRaises
from flsim.utils.async_trainer.async_staleness_weights import (
    AsyncStalenessWeightConfig,
    StalenessWeight,
    ConstantStalenessWeightConfig,
    ThresholdStalenessWeightConfig,
    PolynomialStalenessWeightConfig,
)
from flsim.utils.tests.helpers.async_weights_test_utils import (
    AsyncStalenessWeightsTestUtils,
)
from hydra.utils import instantiate


class TestAsyncStalenessWeights:
    @pytest.mark.parametrize(
        "staleness_weight_config, staleness_weight_class",
        AsyncStalenessWeightsTestUtils.STALENESS_WEIGHT_TEST_CONFIGS,
    )
    def test_string_conversion(
        self,
        staleness_weight_config: AsyncStalenessWeightConfig,
        staleness_weight_class: StalenessWeight,
    ):
        obj = instantiate(staleness_weight_config)
        assertEqual(obj.__class__, staleness_weight_class)

    @pytest.mark.parametrize(
        "avg_staleness",
        AsyncStalenessWeightsTestUtils.AVG_TEST_STALENESS,
    )
    def test_constant_weight_compute(self, avg_staleness):
        """Test that all constant weight computation works as expected"""
        max_staleness = 10000
        obj = instantiate(ConstantStalenessWeightConfig(avg_staleness=avg_staleness))
        for _i in range(10):
            staleness = np.random.randint(1, max_staleness)
            numerator = AsyncStalenessWeightsTestUtils.get_constant_wt()
            denom = AsyncStalenessWeightsTestUtils.get_constant_wt()
            assertEqual(obj.weight(staleness), numerator / denom)

    @pytest.mark.parametrize(
        "avg_staleness",
        AsyncStalenessWeightsTestUtils.AVG_TEST_STALENESS,
    )
    def test_threshold_weight_compute(self, avg_staleness):
        """Test that threshold weight computation works as expected"""
        max_staleness = 10000
        for _i in range(10):
            cutoff = np.random.randint(1, max_staleness)
            value_after_cutoff = np.random.uniform(low=0.0, high=1.0)
            obj = instantiate(
                ThresholdStalenessWeightConfig(
                    avg_staleness=avg_staleness,
                    cutoff=cutoff,
                    value_after_cutoff=value_after_cutoff,
                )
            )
            staleness = np.random.randint(1, max_staleness)
            numerator = AsyncStalenessWeightsTestUtils.get_threshold_wt(
                staleness=staleness,
                cutoff=cutoff,
                value_after_cutoff=value_after_cutoff,
            )
            denom = AsyncStalenessWeightsTestUtils.get_threshold_wt(
                staleness=avg_staleness,
                cutoff=cutoff,
                value_after_cutoff=value_after_cutoff,
            )
            assertEqual(obj.weight(staleness), numerator / denom)

    @pytest.mark.parametrize(
        "avg_staleness",
        AsyncStalenessWeightsTestUtils.AVG_TEST_STALENESS,
    )
    def test_polynomial_weight_compute(self, avg_staleness):
        """Test that threshold weight computation works as expected"""
        max_staleness = 10000
        for _i in range(10):
            exponent = np.random.uniform(low=0.0, high=1.0)
            obj = instantiate(
                PolynomialStalenessWeightConfig(
                    avg_staleness=avg_staleness, exponent=exponent
                )
            )
            staleness = np.random.randint(1, max_staleness)
            numerator = AsyncStalenessWeightsTestUtils.get_polynomial_wt(
                staleness=staleness, exponent=exponent
            )
            denom = AsyncStalenessWeightsTestUtils.get_polynomial_wt(
                staleness=avg_staleness, exponent=exponent
            )
            assertEqual(obj.weight(staleness), numerator / denom)

    def test_polynomial_weight_zero_exponent(self):
        """For polynomial weight, if exponent is zero, wt=1 regardless of
        staleness or average staleness
        """
        max_staleness = 10000
        # test for 10 random values of staleness and avg staleness
        for _i in range(10):
            staleness = np.random.randint(1, max_staleness)
            avg_staleness = np.random.randint(1, max_staleness)
            obj = instantiate(
                PolynomialStalenessWeightConfig(
                    avg_staleness=avg_staleness, exponent=0.0
                )
            )
            assertEqual(obj.weight(staleness), 1.0)

    def test_polynomial_weight_bad_exponent(self):
        """For polynomial weight, exponent must be between 0 and 1, else error"""
        cfg = PolynomialStalenessWeightConfig(avg_staleness=0, exponent=-0.1)

        # negative exponent causes error
        with assertRaises(
            (
                AssertionError,  # with Hydra 1.1
                hydra.errors.HydraException,  # with Hydra 1.0
            ),
        ):
            cfg.exponent = -0.1
            instantiate(cfg)

        # exponent greater than 1.0 causes error
        with assertRaises(
            (
                AssertionError,  # with Hydra 1.1
                hydra.errors.HydraException,  # with Hydra 1.0
            ),
        ):
            cfg.exponent = 1.1
            instantiate(cfg)

        # exponent = 0.0 is fine
        cfg.exponent = 0.0
        instantiate(cfg)

        # exponent = 1.0 is fine
        cfg.exponent = 1.0
        instantiate(cfg)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
import copy
import math
from dataclasses import dataclass
from typing import List

import numpy as np
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
    DurationDistributionConfig,
)
from hydra.utils import instantiate
from omegaconf import MISSING
from omegaconf import OmegaConf


@dataclass
class EventTimingInfo:
    r"""
    Used only for testing
    """
    prev_event_start_to_current_start: int
    duration: int


class IAsyncTrainingStartTimeDistr(abc.ABC):
    """Abstract class for generating training-start events in AsyncFL"""

    def __init__(self, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=AsyncTrainingStartTimeDistrConfig,
            **kwargs,
        )
        assert (
            # pyre-fixme[16]: `IAsyncTrainingStartTimeDistr` has no attribute `cfg`.
            self.cfg.training_rate
            > 0
        ), f"Event rate must be positive, got {self.cfg.training_rate}"

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def time_to_next_event_start(self) -> float:
        pass


class PoissonAsyncTrainingStartTimeDistr(IAsyncTrainingStartTimeDistr):
    """Training start times are poisson distributed"""

    def __init__(self, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PoissonAsyncTrainingStartTimeDistrConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def time_to_next_event_start(self) -> float:
        # if events are poisson distributed,
        # time to next event is exponentially distributed
        # -ln(U)/lambda = time to next event (from CDF of exponential distribution)
        u = np.random.random()
        # pyre-fixme[16]: `PoissonAsyncTrainingStartTimeDistr` has no attribute `cfg`.
        return -(math.log(u)) / self.cfg.training_rate


class ConstantAsyncTrainingStartTimeDistr(IAsyncTrainingStartTimeDistr):
    """Gap between training start times is constant"""

    def __init__(self, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ConstantAsyncTrainingStartTimeDistrConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def time_to_next_event_start(self) -> float:
        # pyre-fixme[16]: `ConstantAsyncTrainingStartTimeDistr` has no attribute `cfg`.
        return 1.0 / self.cfg.training_rate


class IEventGenerator(abc.ABC):
    """Class that generates both training_start and training_duration events"""

    @abc.abstractmethod
    def time_to_next_event_start(self) -> float:
        pass

    @abc.abstractmethod
    def training_duration(self, num_training_examples: int) -> float:
        pass


class AsyncTrainingEventGenerator(IEventGenerator):
    """Class that generates both training_start and training_duration events"""

    def __init__(self, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=AsyncTrainingEventGeneratorConfig,
            **kwargs,
        )
        self._validate_cfg()

        self._training_start_time_distr = instantiate(
            # pyre-fixme[16]: `AsyncTrainingEventGenerator` has no attribute `cfg`.
            self.cfg.training_start_time_distribution
        )
        self._training_duration_distr = instantiate(
            self.cfg.duration_distribution_generator
        )

    def _validate_cfg(self):
        # looping over the config fields throws incase of missing field
        for _ in self.cfg.items():
            pass

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.training_start_time_distribution, "_target_"):
            cfg.training_start_time_distribution = (
                ConstantAsyncTrainingStartTimeDistrConfig()
            )
        if OmegaConf.is_missing(cfg.duration_distribution_generator, "_target_"):
            cfg.duration_distribution_generator = (
                PerExampleGaussianDurationDistributionConfig()
            )

    def time_to_next_event_start(self) -> float:
        return self._training_start_time_distr.time_to_next_event_start()

    def training_duration(self, num_training_examples: int) -> float:
        return self._training_duration_distr.training_duration(num_training_examples)


class AsyncTrainingEventGeneratorFromList(IEventGenerator):
    """This class simulates TrainingEventGenerator
    It returns time-to-next-event and event-duration from a fixed list
    Useful for writing unit tests for components that use TrainingEventGenerator
    """

    def __init__(self, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=AsyncTrainingEventGeneratorFromListConfig,
            **kwargs,
        )
        self._validate_cfg()

        self.distr: List[EventTimingInfo] = copy.deepcopy(
            # pyre-fixme[16]: `AsyncTrainingEventGeneratorFromList` has no attribute
            #  `cfg`.
            list(self.cfg.training_events)
        )
        self.training_events = list(self.cfg.training_events)
        self.current_event: EventTimingInfo = EventTimingInfo(
            prev_event_start_to_current_start=0, duration=0
        )

    def _validate_cfg(self):
        # looping over the config fields throws incase of missing field
        for _ in self.cfg.items():
            pass

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def time_to_next_event_start(self) -> float:
        self.set_next_event()
        return self.current_event.prev_event_start_to_current_start

    def training_duration(self, num_training_examples: int) -> float:
        return self.current_event.duration

    def set_next_event(self):
        if len(self.distr) == 0:
            self.distr = copy.deepcopy(self.training_events)
        self.current_event = self.distr.pop(0)


@dataclass
class AsyncTrainingStartTimeDistrConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    # Average number of devices training per unit time
    training_rate: float = 1.0


@dataclass
class PoissonAsyncTrainingStartTimeDistrConfig(AsyncTrainingStartTimeDistrConfig):
    _target_: str = fullclassname(PoissonAsyncTrainingStartTimeDistr)


@dataclass
class ConstantAsyncTrainingStartTimeDistrConfig(AsyncTrainingStartTimeDistrConfig):
    _target_: str = fullclassname(ConstantAsyncTrainingStartTimeDistr)


@dataclass
class EventGeneratorConfig:
    _target_: str = MISSING
    _recursive_: bool = False


@dataclass
class AsyncTrainingEventGeneratorFromListConfig(EventGeneratorConfig):
    _target_: str = fullclassname(AsyncTrainingEventGeneratorFromList)
    # list of (time-to-next-event-start, event-duration) tuples
    training_events: List[EventTimingInfo] = MISSING


@dataclass
class AsyncTrainingEventGeneratorConfig(EventGeneratorConfig):
    _target_: str = fullclassname(AsyncTrainingEventGenerator)
    training_start_time_distribution: AsyncTrainingStartTimeDistrConfig = (
        AsyncTrainingStartTimeDistrConfig()
    )
    duration_distribution_generator: DurationDistributionConfig = (
        DurationDistributionConfig()
    )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""
This file defines the concept of a base round aggregator for a
federated learning setting. Also defines basic config,
for an FL aggregator. Aggregator is responsible for just efficient
state_dict gathering and potentially later on user-level differential
privacy. It is different from the concept of the server which modifies
updates the global model.
"""

from __future__ import annotations

import abc
import logging
from dataclasses import dataclass
from enum import IntEnum
from itertools import chain
from typing import Optional, Tuple

import torch
from flsim.channels.base_channel import IdentityChannel
from flsim.channels.message import Message
from flsim.common.logger import Logger
from flsim.interfaces.model import IFLModel
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.distributed.fl_distributed import FLDistributedUtils, OperationType
from flsim.utils.fl.common import FLModelParamUtils
from omegaconf import MISSING
from torch import nn


class ReductionType(IntEnum):
    """
    Type of reduction for the RoundReducer
    """

    AVERAGE = 0
    SUM = 1
    WEIGHTED_AVERAGE = 2
    WEIGHTED_SUM = 3


class ReductionPrecision(IntEnum):
    """
    Defines the precision of the aggregated module.

    It can help a little bit if the reduced module has a double precision.
    DEFAULT keeps the precision of the reference module.
    """

    DEFAULT = 0
    FLOAT = 1
    DOUBLE = 2

    @property
    def dtype(self) -> Optional[torch.dtype]:
        return (
            torch.float64
            if self == ReductionPrecision.DOUBLE
            else torch.float32
            if self == ReductionPrecision.FLOAT
            else None  # ReductionPrecision.DEFAULT
        )


class IFLRoundReducer(abc.ABC):
    """
    Interface for RoundReducers.
    """

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=IFLRoundReducerConfig,
            **kwargs,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def collect_update(self, delta: IFLModel) -> None:
        """
        Given a updated model from the client, add it to this reducer.
        """
        pass

    @abc.abstractmethod
    def reduce(self):
        """
        Reduce all the updates collected thus far and return the results.
        """
        pass

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Initializes / Resets round reducers internals.
        """
        pass


@dataclass
class IFLRoundReducerConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    only_federated_params: bool = True


class RoundReducer(IFLRoundReducer):
    """
    Base Class for an aggregator which gets parameters
    from different clients and aggregates them together.
    """

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        *,
        global_model: IFLModel,
        num_users_per_round: Optional[int] = None,
        total_number_of_users: Optional[int] = None,
        channel: Optional[IdentityChannel] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=RoundReducerConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

        # pyre-fixme[16]: `RoundReducer` has no attribute `cfg`.
        self.dtype = self.cfg.precision.dtype
        self.channel = channel or IdentityChannel()
        self.name = name or "unnamed_aggregator"
        self.num_users_per_round = num_users_per_round
        self.total_number_of_users = total_number_of_users
        # TODO these are specific to mean reducer [this implementation]
        # we will probably need a level of inheritence here and hide
        # these frome the main class.
        self.sum_weights: torch.Tensor = torch.zeros(1)
        self.ref_model: IFLModel = global_model
        self.reduced_module: nn.Module = FLModelParamUtils.clone(
            global_model.fl_get_module(), self.dtype
        )
        self._zero_weights()

    def set_num_total_users(self, num_total_users):
        self.total_number_of_users = num_total_users

    def collect_update(self, delta: IFLModel, weight: float) -> None:
        # 0. Receive delta from client through channel
        delta = self.receive_through_channel(delta)
        # 1. reduce the delta into local state
        self.update_reduced_module(delta.fl_get_module(), weight)

    def _reduce_all(self, op: OperationType = OperationType.SUM_AND_BROADCAST):
        """
        reduce models accross all workers if multi-processing is used.
        reduction type is defined by `config.reduction_type`, see
        `ReductionType`.

        Returns:
            number of the models that have been collected. For weighted
            reductions types returns sum of all model weights.

        Note:
            The weights are sum of weights only for weighted reduction types
            see 'ReductionType`, for simple reduction it is the number of models
            that have been reduced.
        """
        state_dict = FLModelParamUtils.get_state_dict(
            self.reduced_module,
            # pyre-fixme[16]: `RoundReducer` has no attribute `cfg`.
            only_federated_params=self.cfg.only_federated_params,
        )
        FLDistributedUtils.distributed_operation(
            chain([self.sum_weights], state_dict.values()), op
        )

        if self.sum_weights.item() <= 0:
            return 0.0
        total_weight = float(self.sum_weights.item())
        if self.is_averaged:
            # reduced_module = reduced_module / total_weight
            FLModelParamUtils.multiply_model_by_weight(
                model=self.reduced_module,
                weight=1 / total_weight,
                model_to_save=self.reduced_module,
                only_federated_params=self.cfg.only_federated_params,
            )

    def receive_through_channel(self, model: IFLModel) -> IFLModel:
        """
        Receives a reference to a state (referred to as model state_dict)
        over the channel. Any channel effect is applied as part of this
        receive function.
        """
        message = self.channel.client_to_server(Message(model))
        return message.model

    @property
    def current_results(self) -> Tuple[nn.Module, float]:
        return self.reduced_module, float(self.sum_weights.item())

    def reduce(self) -> Tuple[nn.Module, float]:
        self._reduce_all()
        return self.reduced_module, float(self.sum_weights.item())

    def reset(self, ref_model: IFLModel) -> None:
        """
        Initializes / Resets round reducers internals.
        """
        self.ref_model = ref_model
        self._zero_weights()

    def _zero_weights(self):
        """
        Reset parameters and weights to zero
        """
        FLModelParamUtils.zero_weights(
            self.reduced_module, only_federated_params=self.cfg.only_federated_params
        )
        device = next(self.reduced_module.parameters()).device
        self.sum_weights = torch.zeros(1, device=device, dtype=self.dtype)

    def update_reduced_module(self, delta_module: nn.Module, weight: float) -> None:
        # TODO num_samples is used as the default weight, this needs revisit
        if not self.is_weighted:
            weight = 1.0
        FLModelParamUtils.linear_comb_models(
            self.reduced_module,
            1.0,
            delta_module,
            weight,
            self.reduced_module,
            # pyre-fixme[16]: `RoundReducer` has no attribute `cfg`.
            only_federated_params=self.cfg.only_federated_params,
        )
        self.sum_weights += weight

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(
                "L1 norm of aggregated parameters:",
                sum(p.abs().sum() for p in self.reduced_module.parameters()),
            )

    @property
    def is_weighted(self):
        return self.cfg.reduction_type in (
            ReductionType.WEIGHTED_SUM,
            ReductionType.WEIGHTED_AVERAGE,
        )

    @property
    def is_averaged(self):
        return self.cfg.reduction_type in (
            ReductionType.WEIGHTED_AVERAGE,
            ReductionType.AVERAGE,
        )


@dataclass
class RoundReducerConfig(IFLRoundReducerConfig):
    _target_: str = fullclassname(RoundReducer)
    reduction_type: ReductionType = ReductionType.WEIGHTED_AVERAGE
    precision: ReductionPrecision = ReductionPrecision.DEFAULT



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import pytest
from flsim.common.pytest_helper import assertEqual, assertTrue
from flsim.data.data_provider import FLDataProviderFromList
from flsim.utils.async_trainer.async_user_selector import (
    RandomAsyncUserSelector,
    RoundRobinAsyncUserSelector,
)
from flsim.utils.sample_model import MockFLModel


@pytest.fixture
def num_users():
    return 20


@pytest.fixture
def num_trials():
    return 100


class TestAsyncUserSelectorUtils:
    def test_random_user_selector(self, num_users, num_trials):
        # users are 0....n-1
        # number of examples per user: [1, 2, 3...., n-1, n]
        num_examples_per_user = list(range(1, num_users + 1))
        data = [
            [1] * num_example
            for num_example, _ in zip(num_examples_per_user, range(num_users))
        ]
        data_provider = FLDataProviderFromList(
            train_user_list=data,
            eval_user_list=data,
            test_user_list=data,
            model=MockFLModel(),
        )
        random_user_selector = RandomAsyncUserSelector(data_provider=data_provider)
        for _ in range(0, num_trials):
            random_user_info = random_user_selector.get_random_user()
            random_user, user_index = (
                random_user_info.user_data,
                random_user_info.user_index,
            )
            assertTrue(user_index >= 0 and user_index < num_users)
            assertEqual(random_user.num_examples(), user_index + 1)

    def test_round_robin_user_selector(self, num_users, num_trials):
        # users are 0....n-1
        # number of examples per user: [10, num_users, 30...., 10*n-1, 10*n]
        multiplier = 10
        num_examples_per_user = [multiplier * i for i in list(range(1, num_users + 1))]
        round_robin_user_selector = RoundRobinAsyncUserSelector(num_examples_per_user)
        data = [
            [1] * num_example
            for num_example, _ in zip(num_examples_per_user, range(num_users))
        ]
        data_provider = FLDataProviderFromList(
            train_user_list=data,
            eval_user_list=data,
            test_user_list=data,
            model=MockFLModel(),
        )
        round_robin_user_selector = RoundRobinAsyncUserSelector(
            data_provider=data_provider
        )
        for num_trial in range(0, num_trials):
            random_user_info = round_robin_user_selector.get_random_user()
            random_user, user_index = (
                random_user_info.user_data,
                random_user_info.user_index,
            )
            assertEqual(user_index, num_trial % num_users)
            assertEqual(random_user.num_examples(), (user_index + 1) * multiplier)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
import copy
import math
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

import torch
from flsim.data.data_provider import IFLDataProvider, IFLUserData
from flsim.interfaces.model import IFLModel
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from omegaconf import MISSING


class ActiveUserSelectorUtils:
    @staticmethod
    def convert_to_probability(
        user_utility: torch.Tensor,
        fraction_with_zero_prob: float,
        softmax_temperature: float,
        weights=None,
    ) -> torch.Tensor:
        if weights is None:
            weights = torch.ones(len(user_utility), dtype=torch.float)
        num_to_zero_out = math.floor(fraction_with_zero_prob * len(user_utility))

        sorted_indices = torch.argsort(user_utility, descending=True).tolist()
        unnormalized_probs = torch.exp(softmax_temperature * user_utility) * weights
        if num_to_zero_out > 0:
            for i in sorted_indices[-num_to_zero_out:]:
                unnormalized_probs[i] = 0

        tmp_sum = sum(unnormalized_probs.tolist())
        assert tmp_sum > 0
        normalized_probs = unnormalized_probs / tmp_sum

        return normalized_probs

    @staticmethod
    def normalize_by_sample_count(
        user_utility: torch.Tensor,
        user_sample_counts: torch.Tensor,
        averaging_exponent: float,
    ) -> torch.Tensor:
        sample_averaging_weights = 1 / torch.pow(user_sample_counts, averaging_exponent)
        user_utility = sample_averaging_weights * user_utility
        return user_utility

    @staticmethod
    def samples_per_user(data_provider: IFLDataProvider) -> torch.Tensor:
        samples_per_user = [
            data_provider.get_user_data(u).num_examples()
            for u in data_provider.user_ids()
        ]
        samples_per_user = torch.tensor(samples_per_user, dtype=torch.float)
        return samples_per_user

    @staticmethod
    def select_users(
        users_per_round: int,
        probs: torch.Tensor,
        fraction_uniformly_random: float,
        rng: Any,
    ) -> List[int]:
        num_total_users = len(probs)
        num_randomly_selected = math.floor(users_per_round * fraction_uniformly_random)
        num_actively_selected = users_per_round - num_randomly_selected
        assert len(torch.nonzero(probs)) >= num_actively_selected

        if num_actively_selected > 0:

            actively_selected_indices = torch.multinomial(
                probs, num_actively_selected, replacement=False, generator=rng
            ).tolist()
        else:
            actively_selected_indices = []

        if num_randomly_selected > 0:
            tmp_probs = torch.tensor(
                [
                    0 if x in actively_selected_indices else 1
                    for x in range(num_total_users)
                ],
                dtype=torch.float,
            )

            randomly_selected_indices = torch.multinomial(
                tmp_probs, num_randomly_selected, replacement=False, generator=rng
            ).tolist()
        else:
            randomly_selected_indices = []

        selected_indices = actively_selected_indices + randomly_selected_indices
        return selected_indices

    @staticmethod
    def sample_available_users(
        users_per_round: int, available_users: List[int], rng: torch.Generator
    ) -> List[int]:
        if users_per_round >= len(available_users):
            return copy.copy(available_users)

        selected_indices = torch.multinomial(
            torch.ones(len(available_users), dtype=torch.float),
            users_per_round,
            replacement=False,
            generator=rng,
        ).tolist()

        return [available_users[idx] for idx in selected_indices]


class ActiveUserSelector(abc.ABC):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ActiveUserSelectorConfig,
            **kwargs,
        )

        self.rng = torch.Generator()
        if self.cfg.user_selector_seed is not None:
            self.rng = self.rng.manual_seed(self.cfg.user_selector_seed)
        else:
            self.rng.seed()

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def get_user_indices(self, **kwargs) -> List[int]:
        pass

    def get_users_unif_rand(
        self, num_total_users: int, users_per_round: int
    ) -> List[int]:
        selected_indices = torch.multinomial(
            torch.ones(num_total_users, dtype=torch.float),
            users_per_round,
            replacement=False,
            generator=self.rng,
        ).tolist()

        return selected_indices

    def unpack_required_inputs(
        self, required_inputs: List[str], kwargs: Dict[str, Any]
    ) -> List[Any]:
        inputs = []
        for key in required_inputs:
            input = kwargs.get(key, None)
            assert (
                input is not None
            ), "Input `{}` is required for get_user_indices in active_user_selector {}.".format(
                key, self.__class__.__name__
            )
            inputs.append(input)
        return inputs


class UniformlyRandomActiveUserSelector(ActiveUserSelector):
    """Simple User Selector which does random sampling of users"""

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=UniformlyRandomActiveUserSelectorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = ["num_total_users", "users_per_round"]
        num_total_users, users_per_round = self.unpack_required_inputs(
            required_inputs, kwargs
        )

        selected_indices = torch.multinomial(
            torch.ones(num_total_users, dtype=torch.float),
            users_per_round,
            # pyre-fixme[16]: `UniformlyRandomActiveUserSelector` has no attribute
            #  `cfg`.
            replacement=self.cfg.random_with_replacement,
            generator=self.rng,
        ).tolist()

        return selected_indices


class SequentialActiveUserSelector(ActiveUserSelector):
    """Simple User Selector which chooses users in sequential manner.
    e.g. if 2 users (user0 and user1) were trained in the previous round,
    next 2 users (user2 and user3) will be picked in the current round.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=SequentialActiveUserSelectorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)
        self.cur_round_user_index = 0

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = ["num_total_users", "users_per_round"]
        num_total_users, users_per_round = self.unpack_required_inputs(
            required_inputs, kwargs
        )

        # when having covered all the users, return the cursor to 0
        if num_total_users <= self.cur_round_user_index:
            self.cur_round_user_index = 0

        next_round_user_index = self.cur_round_user_index + users_per_round
        user_indices = list(
            range(
                self.cur_round_user_index, min(next_round_user_index, num_total_users)
            )
        )
        self.cur_round_user_index = next_round_user_index
        return user_indices


class RandomRoundRobinActiveUserSelector(ActiveUserSelector):
    """User Selector which chooses users randomly in a round-robin fashion.
    Each round users are selected uniformly randomly from the users not
    yet selected in that epoch.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=RandomRoundRobinActiveUserSelectorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)
        self.available_users = []

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = ["num_total_users", "users_per_round"]
        num_total_users, users_per_round = self.unpack_required_inputs(
            required_inputs, kwargs
        )
        # when having covered all the users, reset the list of available users
        if len(self.available_users) == 0:
            self.available_users = list(range(num_total_users))

        user_indices = ActiveUserSelectorUtils.sample_available_users(
            users_per_round, self.available_users, self.rng
        )

        # Update the list of available users
        # TODO(dlazar): ensure this is the fastest method. If not, write a util
        self.available_users = [
            idx for idx in self.available_users if idx not in user_indices
        ]

        return user_indices


class NumberOfSamplesActiveUserSelector(ActiveUserSelector):
    """Active User Selector which chooses users with probability weights determined
    by the number of training points on the client. Clients with more samples are
    given higher probability of being selected. Assumes that number of samples
    on each client is known by the server and updated constantly.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=NumberOfSamplesActiveUserSelectorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = ["users_per_round", "data_provider"]
        users_per_round, data_provider = self.unpack_required_inputs(
            required_inputs, kwargs
        )

        user_utility = ActiveUserSelectorUtils.samples_per_user(data_provider)
        # pyre-fixme[16]: `NumberOfSamplesActiveUserSelector` has no attribute `cfg`.
        user_utility = torch.pow(user_utility, self.cfg.exponent)
        probs = ActiveUserSelectorUtils.convert_to_probability(
            user_utility=torch.log(user_utility),
            fraction_with_zero_prob=self.cfg.fraction_with_zero_prob,
            softmax_temperature=1,
        )
        selected_indices = ActiveUserSelectorUtils.select_users(
            users_per_round=users_per_round,
            probs=probs,
            fraction_uniformly_random=self.cfg.fraction_uniformly_random,
            rng=self.rng,
        )

        return selected_indices


class HighLossActiveUserSelector(ActiveUserSelector):
    """Active User Selector which chooses users with probability weights determined
    by the loss the model suffers on the user's data. Since this is a function of
    both the data on the client and the model, user loss values are not updated
    for every user before a selection round. Instead the class will keep a record
    of the loss and only update the value for a user when they are used for training
    since in this case the model is being transmitted to the user anyway.

    If this staleness become a significant problem, try using ideas from
    non-stationary UCB: https://arxiv.org/pdf/0805.3415.pdf.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=HighLossActiveUserSelectorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

        self.user_losses: torch.Tensor = torch.tensor([], dtype=torch.float)
        self.user_sample_counts: torch.Tensor = torch.tensor([], dtype=torch.float)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _get_initial_losses_and_counts(
        self,
        num_total_users: int,
        data_provider: IFLDataProvider,
        global_model: IFLModel,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        user_losses = torch.zeros(num_total_users, dtype=torch.float)
        user_sample_counts = torch.zeros(num_total_users, dtype=torch.float)
        for i in range(num_total_users):
            (
                user_losses[i],
                user_sample_counts[i],
            ) = self._get_user_loss_and_sample_count(
                data_provider.get_user_data(i), global_model
            )
        return user_losses, user_sample_counts

    def _non_active_sampling(
        self, num_total_users: int, users_per_round: int
    ) -> List[int]:
        selected_indices = torch.multinomial(
            torch.ones(num_total_users, dtype=torch.float),
            users_per_round,
            replacement=False,
            generator=self.rng,
        ).tolist()
        return selected_indices

    @staticmethod
    def _get_user_loss_and_sample_count(
        user_data: IFLUserData, model: IFLModel
    ) -> Tuple[float, int]:
        loss = 0
        num_samples = 0
        for batch in user_data:
            metrics = model.get_eval_metrics(batch)
            loss += metrics.loss.item() * metrics.num_examples
            num_samples += metrics.num_examples
        return loss, num_samples

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = [
            "num_total_users",
            "users_per_round",
            "data_provider",
            "global_model",
            "epoch",
        ]
        (
            num_total_users,
            users_per_round,
            data_provider,
            global_model,
            epoch,
        ) = self.unpack_required_inputs(required_inputs, kwargs)

        # pyre-fixme[16]: `HighLossActiveUserSelector` has no attribute `cfg`.
        if epoch < self.cfg.epochs_before_active:
            selected_indices = self._non_active_sampling(
                num_total_users, users_per_round
            )
            return selected_indices

        if self.user_losses.nelement() == 0:
            (
                self.user_losses,
                self.user_sample_counts,
            ) = self._get_initial_losses_and_counts(
                num_total_users, data_provider, global_model
            )

        user_utility = ActiveUserSelectorUtils.normalize_by_sample_count(
            user_utility=self.user_losses,
            user_sample_counts=self.user_sample_counts,
            averaging_exponent=self.cfg.count_normalization_exponent,
        )
        probs = ActiveUserSelectorUtils.convert_to_probability(
            user_utility=user_utility,
            fraction_with_zero_prob=self.cfg.fraction_with_zero_prob,
            softmax_temperature=self.cfg.softmax_temperature,
        )
        selected_indices = ActiveUserSelectorUtils.select_users(
            users_per_round=users_per_round,
            probs=probs,
            fraction_uniformly_random=self.cfg.fraction_uniformly_random,
            rng=self.rng,
        )

        for i in selected_indices:
            (
                self.user_losses[i],
                self.user_sample_counts[i],
            ) = self._get_user_loss_and_sample_count(
                data_provider.get_user_data(i), global_model
            )

        return selected_indices


@dataclass
class ActiveUserSelectorConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    user_selector_seed: Optional[int] = None


@dataclass
class UniformlyRandomActiveUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(UniformlyRandomActiveUserSelector)
    random_with_replacement: bool = False


@dataclass
class SequentialActiveUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(SequentialActiveUserSelector)


@dataclass
class RandomRoundRobinActiveUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(RandomRoundRobinActiveUserSelector)


@dataclass
class NumberOfSamplesActiveUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(NumberOfSamplesActiveUserSelector)
    exponent: float = 1.0
    fraction_uniformly_random: float = 0.0
    fraction_with_zero_prob: float = 0.0


@dataclass
class HighLossActiveUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(HighLossActiveUserSelector)
    count_normalization_exponent: float = 0.0
    epochs_before_active: int = 0
    fraction_uniformly_random: float = 0.0
    fraction_with_zero_prob: float = 0.0
    softmax_temperature: float = 1.0



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import numpy as np
import pandas as pd
from flsim.utils.fl.stats import (
    ModelSequenceNumberTracker,
    RandomVariableStatsTracker,
    RandomVariableStatsTrackerMA,
    AverageType,
)
from libfb.py import testutil


class StatsTrackerTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_stats_tracker(self) -> None:
        """Test that we can accurately keep track of stats"""
        np.random.seed(100)
        # check mean and standard_deviation using a normal random
        stats_tracker = RandomVariableStatsTracker()
        for _ in range(1000):
            # test with normal random
            rv = np.random.normal(loc=0, scale=10)
            stats_tracker.update(rv)
        self.assertAlmostEqual(stats_tracker.mean(), 0.0, delta=0.5)
        self.assertAlmostEqual(stats_tracker.standard_deviation(), 10.0, delta=1.0)

        stats_tracker2 = RandomVariableStatsTracker()
        for i in range(1000):
            stats_tracker2.update(i - 10)
        self.assertEqual(stats_tracker2.min_val, -10)
        self.assertEqual(stats_tracker2.max_val, 989)

    def test_sequence_tracker(self) -> None:
        seqnum_tracker = ModelSequenceNumberTracker()
        num_global_step = 5
        for _ in range(num_global_step):
            seqnum_tracker.increment()

        self.assertEqual(num_global_step, seqnum_tracker.current_seqnum)

        num_clients = 100
        client_seqnums = np.random.randint(10000, size=num_clients)
        staleness_weights = []
        for client_seqnum in client_seqnums:
            staleness = seqnum_tracker.get_staleness_and_update_stats(
                client_seqnum=client_seqnum
            )
            self.assertEqual(num_global_step - client_seqnum, staleness)
            staleness_weights.append(staleness)

        expected_mean = np.mean(staleness_weights)
        self.assertAlmostEqual(expected_mean, seqnum_tracker.mean(), delta=1e-6)

        expected_sd = np.std(staleness_weights)
        self.assertAlmostEqual(
            expected_sd, seqnum_tracker.standard_deviation(), delta=1e-6
        )

    @testutil.data_provider(
        lambda: (
            {"max_val": 10, "window_size": 5, "average_type": AverageType.SMA},
            {"max_val": 100, "window_size": 50, "average_type": AverageType.SMA},
            {"max_val": 1000, "window_size": 50, "average_type": AverageType.SMA},
            {"max_val": 10, "window_size": 5, "average_type": AverageType.EMA},
            {"max_val": 100, "window_size": 50, "average_type": AverageType.EMA},
            {"max_val": 1000, "window_size": 50, "average_type": AverageType.EMA},
        )
    )
    def test_moving_average(self, max_val, window_size, average_type) -> None:
        decay_factor = 0.5
        stats_tracker = RandomVariableStatsTrackerMA(
            window_size=window_size, mode=average_type
        )

        values = np.arange(1, max_val, 1)
        for i in values:
            stats_tracker.update(i)

        values = np.array(values[-window_size:])

        if average_type == AverageType.SMA:
            expected_mean = values.mean()
            expected_std = values.std()
        else:
            v = pd.Series(values)
            expected_mean = v.ewm(alpha=decay_factor).mean().iloc[-1]
            expected_std = v.ewm(alpha=decay_factor).std().iloc[-1]

        self.assertEqual(stats_tracker.mean(), expected_mean)
        self.assertEqual(stats_tracker.standard_deviation(), expected_std)

    def test_quantiles_tracker(self):
        stats_tracker = RandomVariableStatsTracker(tracks_quantiles=True)
        values = []
        for i in range(100):
            stats_tracker.update(i)
            values.append(i)

        self.assertEqual(stats_tracker.median_val, np.quantile(values, 0.5))
        self.assertEqual(stats_tracker.lower_quartile_val, np.quantile(values, 0.25))
        self.assertEqual(stats_tracker.upper_quartile_val, np.quantile(values, 0.75))



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from collections import OrderedDict
from copy import deepcopy
from typing import Optional

import torch
from flsim.interfaces.model import IFLModel

# there are some pyre errors here related to torch.Tensor operations, please ignore
# them as they work fine despite the error saying otherwise.
class CountSketch:
    """
    Implementation of the CountSketch data structure described here:
    http://dimacs.rutgers.edu/~graham/ssbd/ssbd3.pdf

    CountSketch is a data structure that can used to be compress a series of
    numbers and then decompress them using a fix-sized matrix and a set
    of pairwise independent hash functions.

    This version is designed to compressed IFLModels, where each weight
    simply gets an unique id, which is simply the number of weights added before
    this weight, and the model's parameter names and weight
    tensor shapes are stored for decompression.
    """

    # TODO: Add device: T95215140
    # TODO: Add message to constructor? T95215260
    def __init__(
        self,
        width: int = 10000,
        depth: int = 11,
        prime: int = 2 ** 31 - 1,
        independence: int = 2,
        h: Optional[torch.Tensor] = None,
        g: Optional[torch.Tensor] = None,
        device="cpu",
    ):
        self.width: int = width
        self.depth: int = depth
        self.prime = prime
        self.buckets = torch.zeros((self.depth, self.width), device=device)
        self.independence = 4 if independence == 4 else 2
        self.device = device

        if h is None:
            self.h = torch.randint(
                low=1,
                high=self.prime,
                size=(self.depth, self.independence),
                device=self.device,
            )
        else:
            if list(h.size()) != [self.depth, self.independence]:
                raise AssertionError(
                    f"Hash function h should be of size {[self.depth, self.independence]}, but got {list(h.size())}"
                )
            self.h = h.to(device=self.device)

        if g is None:
            self.g = torch.randint(
                low=1,
                high=self.prime,
                size=(self.depth, self.independence),
                device=self.device,
            )
        else:
            if list(g.size()) != [self.depth, self.independence]:
                raise AssertionError(
                    f"Hash function g should be of size {[self.depth, self.independence]}, but got {list(g.size())}"
                )
            self.g = g.to(device=self.device)

        self.n = 0
        self.param_sizes = OrderedDict()

    # T95859841: Change return shape.
    def compute_hash_vector(self, x: torch.Tensor, hash: torch.Tensor) -> torch.Tensor:
        """
        Computes the hash of a vector x that represents the ids using the hash functions in the parameter hash.

        Args:
            x: the vector of ids, the expected shape is [num_ids]
            hash: the set of the hash functions to use. Expected shape is [self.depth, self.independence]

        Returns:
            Hash values as a torch.Tensor in the size [num_ids, num_hash]
        """

        def pairwise(x: torch.Tensor, a: torch.Tensor, b: torch.Tensor):
            """
            Perform a pairwise hash function that takes advantage of the broadcasting
            rules of tensor operations.

            Args:
                x: the vector of ids, expected shape is [num_ids]
                a: the coefficents used in the hash function,
                    expected shape is [num_ids, self.depth] or [self.depth]
                b: the offset used in the hash function,
                    expected shape is [num_ids, self.depth] or [self.depth]

            Returns:
                A torch.Tensor of size [num_ids, self.depth] that represents the hashes.
            """
            # pyre-ignore
            return (a * x.unsqueeze(-1) + b) % self.prime

        # should actually work for any independence, but need more testing
        # alternative way to do with list comprehension:
        # torch.cat([x ** i for i in range(self.independence)].view(1, -1), dim = 0)
        if self.independence == 4:
            a = hash[:, 0]
            for i in range(1, hash.size(1)):
                b = hash[:, i]
                a = pairwise(x, a, b)
            return a

        return pairwise(x, hash[:, 0], hash[:, 1])

    def h_hash(self, x: torch.Tensor) -> torch.Tensor:
        # pyre-ignore
        return self.compute_hash_vector(x, self.h) % self.width

    def g_hash(self, x: torch.Tensor) -> torch.Tensor:
        # pyre-ignore
        return 2 * (self.compute_hash_vector(x, self.g) % 2) - 1

    def update(self, x: torch.Tensor, weights: torch.Tensor) -> None:
        self.n += x.numel()
        idx = self.h_hash(x)  # [num_id, self.depth]
        sign = self.g_hash(x)  # [num_id, self.depth]
        signed_weights = (sign * weights.view(-1, 1)).t()  # [self.depth, num_id]

        # offset is used because put_ treats self as a 1D tensor.
        offset = (
            torch.arange(0, self.depth, device=self.device).view(-1, 1) * self.width
        )

        # use put_ instead of index_put_ due to better performance on GPU (100x)
        # see N795398 for empirical results.
        self.buckets.put_(idx.t() + offset, signed_weights, accumulate=True)

    def query(self, x: torch.Tensor) -> torch.Tensor:
        idx = self.h_hash(x)
        sign = self.g_hash(x)
        sketched_weights = self.buckets[range(0, self.depth), idx]
        return torch.median(sketched_weights * sign, dim=1)[0]

    def sketch_state_dict(self, state_dict: OrderedDict):
        """
        Sketch a state_dict and all its weights while also resetting params and n.

        Args:
            state_dict: the dictionary containing parameter names and weights, usually
            obtained by calling state_dict() on a nn.Module
        """
        self.reset_buckets()
        self.param_sizes = OrderedDict()
        self.n = 0

        for name, param in state_dict.items():
            self.param_sizes[name] = param.size()
            self.update(
                torch.arange(self.n, self.n + param.numel(), device=self.device),
                param.view(-1),
            )

    def reset_buckets(self):
        self.buckets.fill_(0)

    def set_params(self, state_dict: OrderedDict):
        """
        Sketch a state_dict and all its weights while also resetting params and n.

        Args:
            state_dict: the dictionary containing parameter names and weights, usually
            obtained by calling state_dict() on a nn.Module
        """
        self.reset_buckets()
        self.param_sizes = OrderedDict()
        self.n = 0

        for name, param in state_dict.items():
            self.param_sizes[name] = param.size()
            self.n += param.numel()

    def sketch_model(self, model: IFLModel) -> None:
        self.sketch_state_dict(model.fl_get_module().state_dict())

    def unsketch_model(self, k: int = -1) -> OrderedDict:
        """
        Unsketchs the model by reconstructing the OrderDict of the
        parameters and their weights from self.buckets and self.param_sizes.
        Supports taking the top_k parameters with the largest weights
        and zero out all the other weights.
        """
        if k == -1:
            k = self.n
        elif k > self.n:
            raise AssertionError(
                "Cannot unsketch with a top_k greater than the number of parameters"
            )

        weights = self.query(torch.arange(0, self.n, device=self.device))

        top, indices = torch.topk(torch.abs(weights), k, sorted=True, largest=True)
        mask = torch.zeros_like(weights, device=weights.device)
        mask[indices] = 1
        weights[mask != 1] = 0

        count = 0
        state_dict = OrderedDict()
        for param_name, param_size in self.param_sizes.items():
            state_dict[param_name] = weights[count : count + param_size.numel()].view(
                param_size
            )
            count += param_size.numel()

        return state_dict

    def linear_comb(self, wt1: float, cs, wt2: float):
        self.buckets *= wt1
        self.buckets += cs.buckets * wt2

    # from N778597
    def approx_L1(self):
        estimates = torch.sum(torch.abs(self.buckets), dim=1)
        return torch.median(estimates)

    def approx_L2(self):
        estimates = torch.sum(self.buckets ** 2)
        return torch.sqrt(torch.median(estimates))

    def get_size_in_bytes(self):
        """
        Calculate CountSketch size in bytes.
        """
        return self.buckets.numel() * self.buckets.element_size()

    def to(self, device):
        """
        Moves the CountSketch to device. Up to the user to make sure the device is valid.
        """
        self.buckets = self.buckets.to(device)
        self.h = self.h.to(device)
        self.g = self.g.to(device)
        self.device = device


def clone_count_sketch(copy: CountSketch) -> CountSketch:
    cs = CountSketch(
        copy.width,
        copy.depth,
        copy.prime,
        copy.independence,
        copy.h,
        copy.g,
        copy.device,
    )
    cs.param_sizes = deepcopy(copy.param_sizes)
    cs.buckets = copy.buckets.detach().clone()
    return cs


def linear_comb_count_sketch(
    cs1: CountSketch, wt1: float, cs2: CountSketch, wt2: float
) -> CountSketch:
    cs = clone_count_sketch(cs1)
    cs.buckets = wt1 * cs.buckets + wt2 * cs2.buckets
    return cs



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from typing import Any, List, Dict

import numpy as np
import pytest
import torch
from flsim.common.pytest_helper import (
    assertTrue,
    assertEqual,
    assertLess,
    assertAlmostEqual,
    assertListEqual,
)
from flsim.common.timeline import Timeline
from flsim.common.timeout_simulator import GaussianTimeOutSimulatorConfig
from flsim.interfaces.metrics_reporter import TrainingStage, Channel
from flsim.interfaces.model import IFLModel
from flsim.metrics_reporter.tensorboard_metrics_reporter import FLMetricsReporter
from flsim.optimizers.async_aggregators import (
    AsyncAggregatorConfig,
    FedAdamAsyncAggregatorConfig,
    FedAvgWithLRAsyncAggregatorConfig,
    FedAvgWithLRHybridAggregatorConfig,
)
from flsim.optimizers.optimizer_scheduler import LRBatchSizeNormalizerSchedulerConfig
from flsim.tests.utils import (
    verify_models_equivalent_after_training,
    RandomEvalMetricsReporter,
    MetricsReporterWithMockedChannels,
)
from flsim.utils.async_trainer.async_example_weights import (
    EqualExampleWeightConfig,
    LinearExampleWeightConfig,
)
from flsim.utils.async_trainer.training_event_generator import (
    AsyncTrainingEventGeneratorConfig,
    ConstantAsyncTrainingStartTimeDistrConfig,
    AsyncTrainingEventGeneratorFromListConfig,
    EventTimingInfo,
)
from flsim.utils.sample_model import DummyAlphabetFLModel, ConstantGradientFLModel
from flsim.utils.tests.helpers.async_trainer_test_utils import (
    async_train_one_user,
    create_async_trainer,
    create_event_generator_config,
    get_data,
    get_equal_split_data,
    get_fl_data_provider,
    get_safe_global_lr,
    get_unequal_split_data,
    run_fl_training_with_event_generator,
    run_fl_training,
    assert_fl_nonfl_same,
    assert_fl_nonfl_same_equal_data_split,
)
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
)
from flsim.utils.timing.training_duration_distribution import (
    PerUserUniformDurationDistributionConfig,
)


class TestMetricsReporter(MetricsReporterWithMockedChannels):
    """Don't reset metrics when reported. Else, at the end of
    training, metrics have zeroes
    """

    def reset(self):
        pass


class ConcurrencyMetricsReporter(FLMetricsReporter):
    ACCURACY = "Accuracy"

    def __init__(self, channels: List[Channel]):
        self.concurrency_metrics = []
        self.eval_rounds = []
        super().__init__(channels)

    def compare_metrics(self, eval_metrics, best_metrics):
        print(f"Current eval accuracy: {eval_metrics}%, Best so far: {best_metrics}%")
        if best_metrics is None:
            return True
        return eval_metrics > best_metrics

    def compute_scores(self) -> Dict[str, Any]:
        # compute accuracy
        correct = torch.Tensor([0])
        for i in range(len(self.predictions_list)):
            all_preds = self.predictions_list[i]
            pred = all_preds.data.max(1, keepdim=True)[1]

            assert pred.device == self.targets_list[i].device, (
                f"Pred and targets moved to different devices: "
                f"pred >> {pred.device} vs. targets >> {self.targets_list[i].device}"
            )
            if i == 0:
                correct = correct.to(pred.device)

            correct += pred.eq(self.targets_list[i].data.view_as(pred)).sum()

        # total number of data
        total = sum(len(batch_targets) for batch_targets in self.targets_list)

        accuracy = 100.0 * correct.item() / total

        return {self.ACCURACY: accuracy}

    def create_eval_metrics(
        self, scores: Dict[str, Any], total_loss: float, **kwargs
    ) -> Any:
        return scores[self.ACCURACY]

    def report_metrics(
        self,
        reset,
        stage,
        extra_metrics=None,
        **kwargs,
    ):

        if stage != TrainingStage.EVAL:
            assert (
                extra_metrics is not None
            ), "Async Trainer metrics reporting should have extra metrics"
            metrics = [m for m in extra_metrics if m.name == "Concurrency_Rate"]
            assert (
                len(metrics) == 1
            ), "Conrrency rate should be one of the extra metrics"
            concurrency_rate = metrics[0]
            self.concurrency_metrics.append(concurrency_rate.value)
        else:
            timeline: Timeline = kwargs.get("timeline", Timeline(global_round=1))
            self.eval_rounds.append(timeline.global_round)
        return super().report_metrics(reset, stage, extra_metrics, **kwargs)


class TestAsyncTrainer:

    # TODO: add test_one_user_sequential_user_same
    # TODO: add test where each user has unequal amount of data, both SGD and adam

    def _assert_fl_nonfl_same_unequal_data_split(
        self,
        fl_batch_size: int,
        num_examples: int,
        num_fl_users: int,
        max_examples_per_user: int,
        epochs: int,
        local_lr: float,
        aggregator_config: AsyncAggregatorConfig,
        training_rate=1,
        training_duration_mean=0,
        training_duration_sd=0,
    ):
        torch.manual_seed(1)
        # create dummy FL model on alphabet
        global_model = DummyAlphabetFLModel()
        # if aggregator is FedAdam, each user should have only one batch
        # otherwise non-FL and FL training won't be the same
        one_batch_per_user_only = "FedAdam" in aggregator_config._target_
        fl_data_provider, nonfl_data_loader = get_unequal_split_data(
            num_examples=num_examples,
            num_fl_users=num_fl_users,
            max_examples_per_user=max_examples_per_user,
            fl_batch_size=fl_batch_size,
            model=global_model,
            one_batch_per_user_only=one_batch_per_user_only,
        )
        assertEqual(
            assert_fl_nonfl_same(
                global_model=global_model,
                fl_data_provider=fl_data_provider,
                nonfl_data_loader=nonfl_data_loader,
                epochs=epochs,
                local_lr=local_lr,
                aggregator_config=aggregator_config,
                training_rate=training_rate,
                training_duration_mean=training_duration_mean,
                training_duration_sd=training_duration_sd,
            ),
            "",
        )

    def assert_fl_nonfl_same_one_user_sgd(self):
        """
        Given:
            data_for_fl={user1:batchA, batchB, batchC, ...}
            data_for_non_fl={batchA, batchB, batchC, ...}

        Check that the following produce the same trained model:
        1. FL training, global_opt=SGD, local_lr=x, global_lr=x
        2. Non-FL training, opt=SGD, lr=x
        """
        # try 3 different batch sizes:
        # 4=> user has multiple batches
        # 32=> user has exactly 1 batch, same as #examples
        # 128=> user has exactly 1 batch, but bigger than #examples
        num_examples = 32
        for fl_batch_size in [4, 32, 128]:
            # choose random learning rate between 0 and 10
            local_lr = np.random.random_sample() * 10
            global_lr = get_safe_global_lr(
                fl_batch_size=fl_batch_size, max_examples_per_user=num_examples
            )
            assertEqual(
                assert_fl_nonfl_same_equal_data_split(
                    num_examples=num_examples,
                    num_fl_users=1,
                    fl_batch_size=fl_batch_size,
                    training_rate=1,
                    training_duration_mean=0,
                    training_duration_sd=0,
                    epochs=5,
                    local_lr=local_lr,
                    aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
                ),
                "",
            )

    def assert_fl_nonfl_same_multiple_users_sgd(self):
        """
        Given:
            data_for_fl={user1:batch1A, batch1B..., user2: batch2A, batch2B,...}
            data_for_non_fl={batch1A, batch1B, ..., batch2A, batch2B, ...}

        Check that the following produce the same trained model:
        1. AsyncFL training, training_rate=1 (sequential), global_opt=SGD, local_lr=x, global_lr=x
        2. Non-FL training, opt=SGD, lr=x
        """
        num_fl_users = 4
        num_examples = 32
        # equal data split
        # try 3 different batch size
        # 4=> each user has multiple batches
        # 16=> each user has exactly 1 batch, same as #examples_per_user
        # 128=> each user has exactly 1 batch, but bigger than #examples_per_user
        equal_split_examples_per_user = num_examples / num_fl_users
        for fl_batch_size in [4, 16, 32]:
            # random learning rate between 0 and 10
            local_lr = np.random.random_sample() * 10
            global_lr = get_safe_global_lr(
                fl_batch_size=fl_batch_size,
                max_examples_per_user=equal_split_examples_per_user,
            )
            assertEqual(
                assert_fl_nonfl_same_equal_data_split(
                    num_examples=num_examples,
                    num_fl_users=num_fl_users,
                    fl_batch_size=fl_batch_size,
                    training_rate=1,
                    training_duration_mean=0,
                    training_duration_sd=0,
                    epochs=5,
                    local_lr=local_lr,
                    aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
                ),
                "",
            )

        # unequal data split
        # user1 = 24 examples, user2 = 8 examples
        # 4=> each user has multiple batches
        # 8=> user1 has multiple batches, user 2 has 1 batch
        # cannot try batch size larger than 8
        unequal_split_max_examples_per_user = 24
        for fl_batch_size in [4, 8]:
            # random learning rate between 0 and 10
            local_lr = np.random.random_sample() * 10
            global_lr = get_safe_global_lr(
                fl_batch_size=fl_batch_size,
                max_examples_per_user=unequal_split_max_examples_per_user,
            )
            self._assert_fl_nonfl_same_unequal_data_split(
                num_examples=num_examples,
                num_fl_users=2,
                max_examples_per_user=unequal_split_max_examples_per_user,
                fl_batch_size=fl_batch_size,
                training_rate=1,
                training_duration_mean=0,
                training_duration_sd=0,
                epochs=5,
                local_lr=local_lr,
                aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
            )

    def assert_fl_nonfl_same_one_user_adam(self):
        """
        Given:
            data_for_fl={user1:batchA, batchB, batchC, ...}
            data_for_non_fl={batchA, batchB, batchC, ...}

        Check that the following produce the same trained model:
        1. FL training, global_opt=Adam, local_lr=x, global_lr=x
        2. Non-FL training, opt=Adam, lr=x
        """
        # for Adam, for FL and non-FL to be equivalent,
        # each user must have maximum one batch of data,
        # because local optimizer on each user has to be SGD
        # 32=> user has exactly 1 batch, same as #examples_per_user
        # 128=> user has exactly 1 batch, but bigger than #examples_per_user
        for fl_batch_size in [32, 128]:
            # random learning rate between 0 and 0.01
            global_lr = np.random.random_sample() * 0.01
            assertEqual(
                assert_fl_nonfl_same_equal_data_split(
                    num_examples=32,
                    num_fl_users=1,
                    fl_batch_size=fl_batch_size,
                    training_rate=1,
                    training_duration_mean=0,
                    training_duration_sd=0,
                    epochs=5,
                    local_lr=1.0,
                    aggregator_config=FedAdamAsyncAggregatorConfig(
                        lr=global_lr, eps=1e-2
                    ),
                ),
                "",
            )

    def assert_fl_nonfl_same_multiple_users_adam(self):
        """
        Given:
            data_for_fl={user1:batch1A, batch1B..., user2: batch2A, batch2B,...}
            data_for_non_fl={batch1A, batch1B, ..., batch2A, batch2B, ...}

        Check that the following produce the same trained model:
        1. AsyncFL training, training_rate=1 (sequential), global_opt=Adam, local_lr=x, global_lr=x
        2. Non-FL training, opt=Adam, lr=x
        """
        # for Adam, for FL and non-FL to be equivalent,
        # each user must have maximum one batch of data,
        # because local optimizer on each user has to be SGD
        # 8=> each user has exactly 1 batch, same as #examples_per_user
        # 128=> each user has exactly 1 batch, but bigger than #examples_per_user
        for fl_batch_size in [8, 128]:
            # random learning rate between 0 and 0.001
            global_lr = np.random.random_sample() * 0.001
            assertEqual(
                assert_fl_nonfl_same_equal_data_split(
                    num_examples=32,
                    num_fl_users=4,
                    fl_batch_size=fl_batch_size,
                    training_rate=1,
                    training_duration_mean=0,
                    training_duration_sd=0,
                    epochs=5,
                    local_lr=1.0,
                    aggregator_config=FedAdamAsyncAggregatorConfig(
                        lr=global_lr, eps=1e-2
                    ),
                ),
                "",
            )

    def _test_local_lr_normalization(self, num_users: int):
        """Run training for two tasks:
            Assumption: batch_size > data on any user (so all users have incomplete batches)
            AsyncFLTask1: data = {user1:data1, user2:data2},
                local_lr_normalization = False, weighting scheme = linear, lr = base_lr
            AsyncFLTask2: data = {user1:data1, user2:data2},
                local_lr_normalization = True, weighting scheme = equal, lr = base_lr
            Task1, Step 1:
               LR = base_lr
               Grad = (grad_user1*base_lr)
               Wt = linear = num_ex_user1
               Model_new = Model_old - (grad_user1*base_lr) * num_ex_user1
            Task2, Step 1:
               LR = base_lr * batch_size
               Grad = grad_user1 *(num_ex_user_1/batch_size) * (base_lr*batch_size)
               Wt = equal = 1
               Model_new = Model_old - (grad_user1*num_ex_user1*base_lr)
            So both models should take the same step
        Verify that the trained models have the same parameters
        """
        fl_batch_size = 128
        base_lr = 0.1
        # local_lr_normalization scales learning rate by num_examples_in_batch
        # we upweight the learning rate so that after local_lr_normalization, the
        # fl-task1 has the same learning rate as fl-task2
        upweighted_lr = base_lr * fl_batch_size

        num_total_examples = 32
        epochs = 5
        # create dummy FL model on alphabet
        torch.manual_seed(1)
        init_model = DummyAlphabetFLModel()

        fl_data_provider, _ = get_equal_split_data(
            num_examples=num_total_examples,
            num_fl_users=num_users,
            fl_batch_size=fl_batch_size,
            model=init_model,
        )

        def get_base_optimizer_and_trained_model():
            init_model_local = copy.deepcopy(init_model)
            async_trainer = create_async_trainer(
                model=init_model_local,
                local_lr=base_lr,
                epochs=epochs,
                aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=1.0),
                example_weight_config=LinearExampleWeightConfig(),
                event_generator_config=create_event_generator_config(
                    training_rate=1.0, training_duration_mean=0, training_duration_sd=0
                ),
            )
            trained_fl_model, _ = async_trainer.train(
                data_provider=fl_data_provider,
                metric_reporter=MetricsReporterWithMockedChannels(),
                num_total_users=num_users,
                distributed_world_size=1,
            )
            return trained_fl_model

        def get_lrnorm_optimizer_and_trained_model():
            init_model_local = copy.deepcopy(init_model)
            async_trainer = create_async_trainer(
                model=init_model_local,
                local_lr=upweighted_lr,
                epochs=epochs,
                aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=1.0),
                example_weight_config=EqualExampleWeightConfig(),
                local_lr_scheduler_config=LRBatchSizeNormalizerSchedulerConfig(
                    base_lr=upweighted_lr, local_lr_normalizer=fl_batch_size
                ),
                event_generator_config=create_event_generator_config(
                    training_rate=1.0, training_duration_mean=0, training_duration_sd=0
                ),
            )
            trained_fl_model, _ = async_trainer.train(
                data_provider=fl_data_provider,
                metric_reporter=MetricsReporterWithMockedChannels(),
                num_total_users=num_users,
                distributed_world_size=1,
            )
            return trained_fl_model

        base_trained_model = get_base_optimizer_and_trained_model()
        lrnorm_trained_model = get_lrnorm_optimizer_and_trained_model()
        assertEqual(
            verify_models_equivalent_after_training(
                base_trained_model,
                lrnorm_trained_model,
                init_model,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_local_lr_normalization_single_user(self):
        self._test_local_lr_normalization(num_users=1)

    def test_local_lr_normalization_multiple_users(self):
        self._test_local_lr_normalization(num_users=4)

    def test_async_training_metrics_reporting(self):
        """Verify that in async training, metrics are only reported
        training_end event fires, not when training_start event fires.
        We do this by:
        a) Creating data where user1=8 examples, user2=1 example
        b) Creating training event order as following:
             t1: user1 starts training
             t2: user2 starts training
             t3: user2 finishes training (should cause metric reporting with #examples=1)
             t11: user1 finishes trainig (should cause metric reporting with #examples=8)
             We verify the order in which metrics were reported.
             metric_reporter.num_examples_list should be = [1, 8]
        b) Creating training event order as following:
             t1: user1 starts training
             t2: user2 starts training
             t3: user1 finishes training (should cause metric reporting with #examples=8)
             t10: user2 finishes trainig (should cause metric reporting with #examples=1)
             We verify the order in which metrics were reported.
             metric_reporter.num_examples_list should be = [8, 1]
        """
        # first entry in tuple: time gap between training start and previous training start
        # second entry in tuple: training duration
        num_examples = 9
        num_fl_users = 2
        max_examples_per_user = 8
        fl_batch_size = 8
        num_epochs = 2
        global_model = DummyAlphabetFLModel()

        fl_data_provider = get_fl_data_provider(
            num_examples,
            num_fl_users,
            max_examples_per_user,
            fl_batch_size,
            global_model,
        )

        def _verify_training(
            user_1_start_time_delta: int,
            user1_training_duration: int,
            user2_start_time_delta: int,
            user2_training_duration: int,
            expected_num_examples: List[int],
        ):
            """Given start time and duration for user1 and user2, verify that that
            num_examples in training has order expected_num_examples
            """
            user1_training_events = EventTimingInfo(
                prev_event_start_to_current_start=user_1_start_time_delta,
                duration=user1_training_duration,
            )
            user2_training_events = EventTimingInfo(
                prev_event_start_to_current_start=user2_start_time_delta,
                duration=user2_training_duration,
            )
            async_trainer = create_async_trainer(
                model=global_model,
                local_lr=1.0,
                epochs=num_epochs,
                aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=1.0),
                event_generator_config=AsyncTrainingEventGeneratorFromListConfig(
                    training_events=[user1_training_events, user2_training_events]
                ),
            )

            metric_reporter = TestMetricsReporter()
            fl_model, _ = async_trainer.train(
                data_provider=fl_data_provider,
                metric_reporter=metric_reporter,
                num_total_users=fl_data_provider.num_users(),
                distributed_world_size=1,
            )
            assertEqual(
                metric_reporter.num_examples_list, expected_num_examples * num_epochs
            )

        # user2 completes training before user1
        _verify_training(
            user_1_start_time_delta=1,
            user1_training_duration=3,
            user2_start_time_delta=1,
            user2_training_duration=1,
            expected_num_examples=[1, 8],
        )

        # user1 completes training before user2
        _verify_training(
            user_1_start_time_delta=1,
            user1_training_duration=1,
            user2_start_time_delta=10,
            user2_training_duration=2,
            expected_num_examples=[8, 1],
        )

    def _run_sequential_training_two_users(
        self,
        initial_model: IFLModel,
        local_lr: int,
        first_user_batches: List[Any],
        second_user_batches: List[Any],
    ) -> IFLModel:
        """Sequential training.
        user1 = user with data: first_user_batches
        user2 = user with data: second_user_batches

        order of events:
        user1 starts training
        user1 finishes training. Global model update
        user2 starts training
        user2 finishes training. Global model update
        """
        global_model = copy.deepcopy(initial_model)
        # sequential training. user1 finishes first. user2 takes user1 trained model, and trains
        for batches in [first_user_batches, second_user_batches]:
            updated_global_model = async_train_one_user(
                global_model_at_training_start=global_model,
                global_model_at_training_end=global_model,
                batches=batches,
                local_lr=local_lr,
            )
            global_model = copy.deepcopy(updated_global_model)
        return global_model

    def _run_parallel_training_two_users(
        self,
        initial_model: IFLModel,
        local_lr: int,
        first_user_batches: List[Any],
        second_user_batches: List[Any],
    ) -> IFLModel:
        """Parallel training.
        user1 = user with data: first_user_batches
        user2 = user with data: second_user_batches

        order of events:
        user1 starts training
        user2 start training
        user1 finishes training. Global model update
        user2 finishes training. Global model update
        NB: this function and _run_sequential_training_two_users share a lot of code, but
            they are much easier to read as separate functions. Trading off code duplication
            for readability.
        """
        global_model = copy.deepcopy(initial_model)
        for batches in [first_user_batches, second_user_batches]:
            # both users start training with the same initial model
            # loop iteration 1 will handle first_user_batches, which will change
            # global_model. so global_model in loop iteration 2 will already include
            # changes from loop iteration 1
            updated_global_model = async_train_one_user(
                global_model_at_training_start=initial_model,
                global_model_at_training_end=global_model,
                batches=batches,
                local_lr=local_lr,
            )
            global_model = copy.deepcopy(updated_global_model)
        return global_model

    def test_num_examples_computation(self):
        r"""
        Test that num_examples for each user affects training duration. Test this for different batch sizes
        Two users: U1 and U2. U2 always has 2 examples. U1 may have 2, 4, or 6 examples
        Training duration = number of examples
        Timing:
          U1 has 2 examples. Sequential training
            t=0 = U1 starts training.
            t=2 = U1 ends training
            t=3 = U2 starts training
            t=5 = U2 ends training
          U1 has 4 examples. U1 and U2 train in parallel. U1 ends first
            t=0 = U1 starts training.
            t=3 = U2 starts training
            t=4 = U1 ends training
            t=5 = U2 ends training
          U1 has 6 examples. U1 and U2 train in parallel. U2 ends first
            t=0 = U1 starts training.
            t=3 = U2 starts training
            t=5 = U2 ends training
            t=6 = U1 ends training
        Verify by comparing with non-FL training that the above timing is correct.
        Do this for batch_size=1 and batch_size=2
        """
        num_epochs = 1
        torch.manual_seed(1)
        initial_model = DummyAlphabetFLModel()
        local_lr = 1.0
        global_lr = 1.0
        num_examples_user2 = 2
        batch_size = 2
        for num_examples_user1 in [2, 4, 6]:
            fl_model = copy.deepcopy(initial_model)
            fl_data_provider, nonfl_data_loader = get_data(
                num_examples=num_examples_user1 + num_examples_user2,
                num_fl_users=2,
                examples_per_user=num_examples_user1,
                fl_batch_size=batch_size,
                nonfl_batch_size=batch_size,
                model=initial_model,
            )
            fl_trained_model = run_fl_training_with_event_generator(
                fl_model=fl_model,
                fl_data_provider=fl_data_provider,
                epochs=num_epochs,
                local_lr=local_lr,
                aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
                training_event_generator_config=create_event_generator_config(
                    training_rate=0.33, training_duration_mean=1, training_duration_sd=0
                ),
            )
            # first num_examples_user1/batch_size batches are for user1. Rest for user2
            all_batches = list(nonfl_data_loader)
            num_user1_batches = num_examples_user1 // batch_size
            user1_batches = all_batches[:num_user1_batches]
            user2_batches = all_batches[num_user1_batches:]
            if num_examples_user1 == 2:
                # U1 training takes 2 units of time
                # sequence:
                # t=0 U1 starts training
                # t=2 U1 finishes training
                # t=3 U2 starts training
                # t=5 U2 finishes training
                simulated_global_model = self._run_sequential_training_two_users(
                    first_user_batches=user1_batches,
                    second_user_batches=user2_batches,
                    initial_model=initial_model,
                    local_lr=local_lr,
                )
            elif num_examples_user1 == 4:
                # U1 training takes 4 units of time
                # sequence:
                # t=0 U1 starts training
                # t=3 U2 starts training
                # t=4 U1 finishes training
                # t=5 U2 finishes training
                simulated_global_model = self._run_parallel_training_two_users(
                    first_user_batches=user1_batches,
                    second_user_batches=user2_batches,
                    initial_model=initial_model,
                    local_lr=local_lr,
                )
            elif num_examples_user1 == 6:
                # U1 training takes 6 units of time
                # sequence:
                # t=0 U1 starts training
                # t=3 U2 starts training
                # t=5 U2 finishes training
                # t=6 U1 finishes training
                simulated_global_model = self._run_parallel_training_two_users(
                    first_user_batches=user2_batches,
                    second_user_batches=user1_batches,
                    initial_model=initial_model,
                    local_lr=local_lr,
                )
            error_msg = verify_models_equivalent_after_training(
                fl_trained_model,
                simulated_global_model,
                initial_model,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            )
            assertEqual(error_msg, "")

    def test_async_training_with_timeout(self):
        """
        Test async training with timeout on.

        Under the condition when mean_per_example=1, std_per_example=0
        The total number of examples trained should be less than the timeout limit

        The test does the following:
        1. Create a data provider with 10 users and each user with 10 examples
        2. Set batch size = 1, hence the metrics reporter losses = num_trained_examples
        3. Check if num_trained_examples <= timeout_limit * num_users
        """
        local_lr = 1.0
        global_lr = 1.0
        epochs = 1
        init_model_local = DummyAlphabetFLModel()
        metrics_reporter = TestMetricsReporter()

        num_users = 10
        num_examples = 100

        timeout_limit = np.random.randint(1, 10)
        async_trainer = create_async_trainer(
            model=init_model_local,
            local_lr=local_lr,
            epochs=epochs,
            aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
            event_generator_config=create_event_generator_config(
                training_rate=1.0, training_duration_mean=0.0, training_duration_sd=0.0
            ),
            timeout_simulator_config=GaussianTimeOutSimulatorConfig(
                timeout_wall_per_round=timeout_limit,
                fl_stopping_time=1e4,
                duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                    training_duration_mean=1.0, training_duration_sd=0.0
                ),
            ),
        )

        data_provider = get_fl_data_provider(
            num_examples=num_examples,
            num_fl_users=num_users,
            examples_per_user=10,
            batch_size=1,
            model=init_model_local,
        )

        final_model, _ = async_trainer.train(
            data_provider=data_provider,
            metric_reporter=metrics_reporter,
            num_total_users=num_users,
            distributed_world_size=1,
        )
        total_examples_trained = len(metrics_reporter.losses)
        total_time = timeout_limit * num_users
        assertTrue(total_examples_trained <= total_time)

    def test_max_staleness_cutoff(self):
        """
        Test for max staleness cut off

        There are two scenario:
        1. Users are trained in parallel with training_mean = 1,
        and max_staleness is some small number then not all users will participate in training
        2. Users are trained sequentially, which means there is no staleness hence all users can participate
        """
        global_model = DummyAlphabetFLModel()
        metrics_reporter = TestMetricsReporter()
        num_users = 10
        num_examples = num_users
        training_rate = num_users
        max_staleness = np.random.randint(1, 5)

        # training in parallel so not all clients can participate
        async_trainer = create_async_trainer(
            model=global_model,
            local_lr=1.0,
            epochs=1,
            aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=1.0),
            event_generator_config=create_event_generator_config(
                training_rate=training_rate,
                training_duration_mean=1.0,
                training_duration_sd=0,
            ),
            max_staleness=max_staleness,
        )

        data_provider = get_fl_data_provider(
            num_examples=num_examples,
            num_fl_users=num_users,
            examples_per_user=1,
            batch_size=1,
            model=global_model,
        )

        final_model, _ = async_trainer.train(
            data_provider=data_provider,
            metric_reporter=metrics_reporter,
            num_total_users=num_users,
            distributed_world_size=1,
        )
        assertLess(async_trainer.global_round, num_users)

        # train sequentially then all clients can participate
        async_trainer = create_async_trainer(
            model=global_model,
            local_lr=1.0,
            epochs=1,
            aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=1.0),
            event_generator_config=create_event_generator_config(
                training_rate=training_rate,
                training_duration_mean=0,
                training_duration_sd=0,
            ),
            max_staleness=0,
        )

        data_provider = get_fl_data_provider(
            num_examples=num_examples,
            num_fl_users=num_users,
            examples_per_user=1,
            batch_size=1,
            model=global_model,
        )

        final_model, _ = async_trainer.train(
            data_provider=data_provider,
            metric_reporter=metrics_reporter,
            num_total_users=num_users,
            distributed_world_size=1,
        )
        assertTrue(async_trainer.global_round - 1 == num_users)

    def test_number_of_steps(self):
        """This test checks that async training takes the same number of optimizer.step()
        as #epochs * #users * #batches_per_user
        """
        num_epochs = 5
        torch.manual_seed(1)
        batch_size = 10
        for num_fl_users in [1, 2]:
            for batches_per_user in [1, 2]:
                fl_model = ConstantGradientFLModel()
                num_total_examples = num_fl_users * batch_size * batches_per_user
                examples_per_user = batch_size * batches_per_user
                fl_data_provider = get_fl_data_provider(
                    num_examples=num_total_examples,
                    num_fl_users=num_fl_users,
                    examples_per_user=examples_per_user,
                    batch_size=batch_size,
                    model=fl_model,
                )
                run_fl_training(
                    fl_model=fl_model,
                    fl_data_provider=fl_data_provider,
                    epochs=num_epochs,
                    local_lr=1.0,
                    aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=1.0),
                    training_rate=1,
                    training_duration_mean=0,
                    training_duration_sd=0,
                )
                # Async FL training is sequential since mean_training_time is 0
                num_optimizer_steps = num_epochs * num_fl_users * batches_per_user
                # ConstantGradientFLModel has a property that its bias term = #of times optimizer.step() is called
                assertTrue(
                    np.isclose(
                        fl_model.fl_get_module().bias.detach().item(),
                        num_optimizer_steps,
                    ),
                    f"Expected: {num_optimizer_steps}, Found: {fl_model.fl_get_module().bias.detach().item()}",
                )

    def test_best_eval_model_is_kept(self):
        """This test checks that AsyncTrainer retains the model with the best eval performance
        To check for this, we use a special MetricsReporter that tracks which model produced
        the best eval results
        """
        num_epochs = 5
        torch.manual_seed(1)
        np.random.seed(1)
        # ConstantGradientFLModel has a property that its bias term = #of times optimizer.step() is called
        batch_size = 10
        num_fl_users = 10
        batches_per_user = 1
        fl_model = ConstantGradientFLModel()
        num_total_examples = num_fl_users * batch_size * batches_per_user
        examples_per_user = batch_size * batches_per_user
        for buffer_size in [1, 2, 3]:
            fl_data_provider = get_fl_data_provider(
                num_examples=num_total_examples,
                num_fl_users=num_fl_users,
                examples_per_user=examples_per_user,
                batch_size=batch_size,
                model=fl_model,
            )
            metric_reporter = RandomEvalMetricsReporter()
            # run_fl_training returns test results. However,
            # RandomEvalMetricsReporter() fakes it so test results are always
            # the best_eval_results
            best_model, best_eval_results = run_fl_training(
                fl_model=fl_model,
                fl_data_provider=fl_data_provider,
                epochs=num_epochs,
                local_lr=1.0,
                aggregator_config=FedAvgWithLRHybridAggregatorConfig(
                    lr=1.0, buffer_size=buffer_size
                ),
                metrics_reporter=metric_reporter,
                do_eval=True,
                report_train_metrics_after_aggregation=True,
                eval_epoch_frequency=0.001,  # report every global model update
            )
            assertEqual(best_eval_results, metric_reporter.best_eval_result)
            # TODO: also check that best_model matches metric_reporter.best_eval_model
            # after fixing code

    @pytest.mark.parametrize(
        "num_users,training_rate,num_epochs", [(100, 10, 2), (50, 10, 2)]
    )
    def test_constant_concurrency(self, num_users, training_rate, num_epochs):
        """
        Test for constant concurrency from one epoch to another
        We expect training_rate #users to be training simultaneously

        In this test, we expect these invariants to hold
        1. Concurrency rate ~ training rate from epoch 1 to epoch n during convergence
        2. Mean converging concurrency for all of the epochs ~ training rate
        3. Async trainer will stop after each epoch to run the eval loop therefore
            we will evaluate every num_users rounds
        """
        torch.manual_seed(2)

        global_model = DummyAlphabetFLModel()
        num_examples = num_users
        event_generator_config = AsyncTrainingEventGeneratorConfig(
            training_start_time_distribution=ConstantAsyncTrainingStartTimeDistrConfig(
                training_rate=training_rate
            ),
            duration_distribution_generator=PerUserUniformDurationDistributionConfig(
                training_duration_mean=1.0,
                training_duration_min=0.01,
            ),
        )
        metrics_reporter = ConcurrencyMetricsReporter([Channel.STDOUT])
        # training in parallel so not all clients can participate
        async_trainer = create_async_trainer(
            model=global_model,
            local_lr=1.0,
            epochs=num_epochs,
            aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=1.0),
            event_generator_config=event_generator_config,
            do_eval=True,
            train_metrics_reported_per_epoch=num_users,
        )

        data_provider = get_fl_data_provider(
            num_examples=num_examples,
            num_fl_users=num_users,
            examples_per_user=1,
            batch_size=1,
            model=global_model,
        )

        final_model, _ = async_trainer.train(
            data_provider=data_provider,
            metric_reporter=metrics_reporter,
            num_total_users=num_users,
        )
        # invariant 1, after check that concurrency eventually reaches
        # steady state and is similar to training rate
        steady_state_seqnum = int(0.9 * len(metrics_reporter.concurrency_metrics))
        for c in metrics_reporter.concurrency_metrics[steady_state_seqnum:]:
            assertAlmostEqual(c, training_rate, delta=1.0)
        # invariant 2
        assertAlmostEqual(
            np.mean(metrics_reporter.concurrency_metrics[steady_state_seqnum:]),
            training_rate,
            delta=1.0,
        )
        # invariant 3
        assertEqual(len(metrics_reporter.eval_rounds), num_epochs)
        assertListEqual(
            metrics_reporter.eval_rounds,
            [(i * num_users) for i in range(1, num_epochs + 1)],
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import numpy as np
import pytest
from flsim.common.pytest_helper import assertEqual
from flsim.utils.async_trainer.async_example_weights import (
    AsyncExampleWeightConfig,
    ExampleWeight,
)
from flsim.utils.tests.helpers.async_weights_test_utils import (  # noqa
    AsyncExampleWeightsTestUtils,
)
from hydra.utils import instantiate


class TestAsyncExampleWeights:
    @pytest.mark.parametrize(
        "example_weight_config, example_weight_class",
        AsyncExampleWeightsTestUtils.EXAMPLE_WEIGHT_TEST_CONFIGS,
    )
    def test_string_conversion(
        self,
        example_weight_config: AsyncExampleWeightConfig,
        example_weight_class: ExampleWeight,
    ):
        """Check that strings are correctly converted to ExampleWeight"""
        obj = instantiate(example_weight_config)
        assertEqual(obj.__class__, example_weight_class)

    @pytest.mark.parametrize(
        "example_weight_config, example_weight_class",
        AsyncExampleWeightsTestUtils.EXAMPLE_WEIGHT_TEST_CONFIGS,
    )
    @pytest.mark.parametrize(
        "avg_num_examples",
        AsyncExampleWeightsTestUtils.AVG_NUMBER_OF_EXAMPLES,
    )
    def test_example_weight_compute(
        self,
        example_weight_config: AsyncExampleWeightConfig,
        example_weight_class: ExampleWeight,
        avg_num_examples: int,
    ):
        """Test that all weight computation works as expected"""
        # generate 10 random integers
        max_num_examples = 10000
        for _ in range(10):
            num_examples = np.random.randint(1, max_num_examples)
            example_weight_config.avg_num_examples = avg_num_examples
            obj = instantiate(example_weight_config)
            assertEqual(
                obj.weight(num_examples),
                AsyncExampleWeightsTestUtils.expected_weight(
                    avg_num_examples=avg_num_examples,
                    num_examples=num_examples,
                    example_weight_class=example_weight_class,
                ),
            )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from dataclasses import dataclass, field
from typing import OrderedDict, Optional

import torch.nn as nn
from flsim.interfaces.model import IFLModel
from flsim.utils.count_sketch import CountSketch
from torch import Tensor


@dataclass
class Message:
    """
    Generic message dataclass, composed of:
        - model: a model containing information that will be sent
        - meta: any meta information about a client or a server
          for instance.

    This dataclass can be extended to your custom needs see the
    ``Message` example.

    Notes:
      - The packet may change after transmission through a channel. For
        instance, the SketchChannel takes as input a model state dict
        and outputs a CountSketch (but no model state dict).
    """

    # model
    model: IFLModel = field(default_factory=nn.Module)

    # add any meta information here
    weight: float = field(default_factory=float)

    # here we store state dict for conveninence
    model_state_dict: OrderedDict[str, Tensor] = field(
        default_factory=OrderedDict[str, Tensor], init=True
    )
    # count sketch
    count_sketch: Optional[CountSketch] = field(default=None)

    def populate_state_dict(self, **kwargs):
        """
        We copy the model's state dict and add is as attribute to the message.

        Notes:
          - We deepcopy the state dict to avoid side effects in case we manipulate
            the state dict inplace.
          - We rely on a model's state dict as it will be easier to change the
            type of the underlying tensors (say int8) versus replacing every
            nn.Module with its corresponding counterpart.
        """

        self.model_state_dict = deepcopy(self.model.fl_get_module().state_dict())

    def update_model_(self):
        """
        Updates model with the state dict stored in the message. May be useful
        when receiving a `Message` and wanting to update the local model.
        """
        assert (
            self.model_state_dict
        ), "Message state dict is empty. Please check if message.state_dict is populated."
        self.model.fl_get_module().load_state_dict(self.model_state_dict)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Any

import torch
from flsim.interfaces.batch_metrics import IFLBatchMetrics


class FLBatchMetrics(IFLBatchMetrics):
    def __init__(
        self,
        *,
        loss: torch.Tensor,
        num_examples: int,
        predictions: torch.Tensor,
        targets: torch.Tensor,
        model_inputs: Any,
    ) -> None:
        self._loss = loss
        self._num_examples = num_examples
        self._predictions = predictions
        self._targets = targets
        self._model_inputs = model_inputs

    @property
    def loss(self) -> torch.Tensor:
        return self._loss

    @property
    def num_examples(self) -> int:
        return self._num_examples

    @property
    def predictions(self) -> torch.Tensor:
        return self._predictions

    @property
    def targets(self) -> torch.Tensor:
        return self._targets

    @property
    def model_inputs(self) -> Any:
        return self._model_inputs
